from numpy.testing import assert_equal
from dipy.align.streamwarp import (bundlewarp,
                                   bundlewarp_shape_analysis,
                                   bundlewarp_vector_filed)
import pytest
from dipy.data import two_cingulum_bundles
from dipy.tracking.streamline import set_number_of_points, Streamlines


def test_bundlewarp():

    cingulum_bundles = two_cingulum_bundles()

    cb1 = Streamlines(cingulum_bundles[0])
    cb1 = set_number_of_points(cb1, 20)

    cb2 = Streamlines(cingulum_bundles[1])
    cb2 = set_number_of_points(cb2, 20)

    deformed_bundle, affine_bundle, dists, mp, warp = bundlewarp(cb1, cb2)

    assert_equal(len(affine_bundle), len(cb2))
    assert_equal(len(deformed_bundle), len(cb2))
    assert_equal(len(deformed_bundle), len(affine_bundle))

    assert_equal(dists.shape, (len(cb2), len(cb1)))

    assert_equal(len(cb2), len(mp))

    assert_equal(len(cb2), len(warp))


def test_bundlewarp_vector_filed():

    cingulum_bundles = two_cingulum_bundles()

    cb1 = Streamlines(cingulum_bundles[0])
    cb1 = set_number_of_points(cb1, 20)

    cb2 = Streamlines(cingulum_bundles[1])
    cb2 = set_number_of_points(cb2, 20)

    deformed_bundle, affine_bundle, dists, mp, warp = bundlewarp(cb1, cb2)

    offsets, directions, colors = bundlewarp_vector_filed(affine_bundle,
                                                          deformed_bundle)

    assert_equal(len(offsets), len(cb2.get_data()))
    assert_equal(len(directions), len(cb2.get_data()))
    assert_equal(len(colors), len(cb2.get_data()))

    assert_equal(len(offsets), len(deformed_bundle.get_data()))
    assert_equal(len(directions), len(deformed_bundle.get_data()))
    assert_equal(len(colors), len(deformed_bundle.get_data()))


def test_bundle_shape_profile():

    cingulum_bundles = two_cingulum_bundles()

    cb1 = Streamlines(cingulum_bundles[0])
    cb1 = set_number_of_points(cb1, 20)

    cb2 = Streamlines(cingulum_bundles[1])
    cb2 = set_number_of_points(cb2, 20)

    deformed_bundle, affine_bundle, dists, mp, warp = bundlewarp(cb1, cb2)

    n = 10
    shape_profile, stdv = bundlewarp_shape_analysis(affine_bundle,
                                                    deformed_bundle,
                                                    no_disks=n)

    assert_equal(len(shape_profile), n)
    assert_equal(len(stdv), n)

    n = 100
    shape_profile, stdv = bundlewarp_shape_analysis(affine_bundle,
                                                    deformed_bundle,
                                                    no_disks=n)

    assert_equal(len(shape_profile), n)
    assert_equal(len(stdv), n)



"""
Diffusion Imaging in Python
============================

For more information, please visit https://dipy.org

Subpackages
-----------
::

 align         -- Registration, streamline alignment, volume resampling
 boots         -- Bootstrapping algorithms
 core          -- Spheres, gradient tables
 core.geometry -- Spherical geometry, coordinate and vector manipulation
 core.meshes   -- Point distributions on the sphere
 data          -- Small testing datasets
 denoise       -- Denoising algorithms
 direction     -- Manage peaks and tracking
 io            -- Loading/saving of dpy datasets
 nn            -- Neural networks algorithms
 reconst       -- Signal reconstruction modules (tensor, spherical harmonics,
                  diffusion spectrum, etc.)
 segment       -- Tractography segmentation
 sims          -- MRI phantom signal simulation
 stats         -- Tractometry
 tracking      -- Tractography, metrics for streamlines
 viz           -- Visualization and GUIs
 workflows      -- Predefined Command line for common tasks

Utilities
---------
::

 test          -- Run unittests
 __version__   -- Dipy version

"""
import sys

from dipy.version import version as __version__
from .testing import setup_test

# Plumb in version etc info stuff
from .pkg_info import get_pkg_info as _get_pkg_info


def get_info():
    from os.path import dirname
    return _get_pkg_info(dirname(__file__))


del sys

submodules = [
    'align',
    'boots',
    'core',
    'data',
    'denoise',
    'direction',
    'io',
    'nn',
    'reconst',
    'segment',
    'sims',
    'stats',
    'tracking',
    'utils',
    'viz',
    'workflows',
    'tests',
    'testing'
]

__all__ = submodules + ['__version__', 'setup_test', 'get_info']



import os.path as op
import pytest
from tempfile import TemporaryDirectory

import numpy as np
import numpy.testing as npt

import nibabel as nib

import dipy.data as dpd
import dipy.core.gradients as dpg

from dipy.align import (syn_registration, register_series, register_dwi_series,
                        center_of_mass, translation, rigid_isoscaling,
                        rigid_scaling, rigid, affine, motion_correction,
                        affine_registration, streamline_registration,
                        write_mapping, read_mapping, register_dwi_to_template)

from dipy.align.imwarp import DiffeomorphicMap

from dipy.tracking.utils import transform_tracking_output
from dipy.io.streamline import save_trk
from dipy.io.stateful_tractogram import StatefulTractogram, Space
from dipy.io.gradients import read_bvals_bvecs
from dipy.io.image import load_nifti
from dipy.testing.decorators import set_random_number_generator


def setup_module():
    global subset_b0, subset_dwi_data, subset_t2, subset_b0_img, \
           subset_t2_img, gtab, hardi_affine, MNI_T2_affine
    MNI_T2 = dpd.read_mni_template()
    hardi_img, gtab = dpd.read_stanford_hardi()
    MNI_T2_data = MNI_T2.get_fdata()
    MNI_T2_affine = MNI_T2.affine
    hardi_data = hardi_img.get_fdata()
    hardi_affine = hardi_img.affine
    b0 = hardi_data[..., gtab.b0s_mask]
    mean_b0 = np.mean(b0, -1)

    # We select some arbitrary chunk of data so this goes quicker:
    subset_b0 = mean_b0[40:45, 40:45, 40:45]
    subset_dwi_data = nib.Nifti1Image(hardi_data[40:45, 40:45, 40:45],
                                      hardi_affine)
    subset_t2 = MNI_T2_data[40:50, 40:50, 40:50]
    subset_b0_img = nib.Nifti1Image(subset_b0, hardi_affine)
    subset_t2_img = nib.Nifti1Image(subset_t2, MNI_T2_affine)


def test_syn_registration():
    with TemporaryDirectory() as tmpdir:
        warped_moving, mapping = syn_registration(subset_b0,
                                                  subset_t2,
                                                  moving_affine=hardi_affine,
                                                  static_affine=MNI_T2_affine,
                                                  step_length=0.1,
                                                  metric='CC',
                                                  dim=3,
                                                  level_iters=[5, 5, 5],
                                                  sigma_diff=2.0,
                                                  radius=1,
                                                  prealign=None)

        npt.assert_equal(warped_moving.shape, subset_t2.shape)
        mapping_fname = op.join(tmpdir, 'mapping.nii.gz')
        write_mapping(mapping, mapping_fname)
        file_mapping = read_mapping(mapping_fname,
                                    subset_b0_img,
                                    subset_t2_img)

        # Test that it has the same effect on the data:
        warped_from_file = file_mapping.transform(subset_b0)
        npt.assert_equal(warped_from_file, warped_moving)

        # Test that it is, attribute by attribute, identical:
        for k in mapping.__dict__:
            npt.assert_((np.all(mapping.__getattribute__(k) ==
                                file_mapping.__getattribute__(k))))


def test_register_dwi_to_template():
    # Default is syn registration:
    warped_b0, mapping = register_dwi_to_template(subset_dwi_data, gtab,
                                                  template=subset_t2_img,
                                                  level_iters=[5, 5, 5],
                                                  sigma_diff=2.0,
                                                  radius=1)
    npt.assert_(isinstance(mapping, DiffeomorphicMap))
    npt.assert_equal(warped_b0.shape, subset_t2_img.shape)

    # Use affine registration (+ don't provide a template and inputs as
    # strings):
    fdata, fbval, fbvec = dpd.get_fnames('small_64D')
    warped_data, affine_mat = register_dwi_to_template(fdata, (fbval, fbvec),
                                                       reg_method="aff",
                                                       level_iters=[5, 5, 5],
                                                       sigmas=[3, 1, 0],
                                                       factors=[4, 2, 1])
    npt.assert_(isinstance(affine_mat, np.ndarray))
    npt.assert_(affine_mat.shape == (4, 4))


def test_affine_registration():
    moving = subset_b0
    static = subset_b0
    moving_affine = static_affine = np.eye(4)
    xformed, affine_mat = affine_registration(moving, static,
                                              moving_affine=moving_affine,
                                              static_affine=static_affine,
                                              level_iters=[5, 5],
                                              sigmas=[3, 1],
                                              factors=[2, 1])

    # We don't ask for much:
    npt.assert_almost_equal(affine_mat[:3, :3], np.eye(3), decimal=1)

    # [center_of_mass] + ret_metric=True should raise an error
    with pytest.raises(ValueError):
        # For array input, must provide affines:
        xformed, affine_mat = affine_registration(moving, static,
                                                  moving_affine=moving_affine,
                                                  static_affine=static_affine,
                                                  pipeline=["center_of_mass"],
                                                  ret_metric=True)

    # Define list of methods
    reg_methods = ["center_of_mass", "translation", "rigid",
                   "rigid_isoscaling", "rigid_scaling", "affine",
                   center_of_mass, translation, rigid,
                   rigid_isoscaling, rigid_scaling, affine]

    # Test methods individually (without returning any metric)
    for func in reg_methods:
        xformed, affine_mat = affine_registration(moving, static,
                                                  moving_affine=moving_affine,
                                                  static_affine=static_affine,
                                                  level_iters=[5, 5],
                                                  sigmas=[3, 1],
                                                  factors=[2, 1],
                                                  pipeline=[func])
        # We don't ask for much:
        npt.assert_almost_equal(affine_mat[:3, :3], np.eye(3), decimal=1)

    # Bad method
    with pytest.raises(ValueError, match=r'^pipeline\[0\] must be one.*foo.*'):
        affine_registration(
            moving, static, moving_affine, static_affine, pipeline=['foo'])

    # Test methods individually (returning quality metric)
    expected_nparams = [0, 3, 6, 7, 9, 12] * 2
    assert len(expected_nparams) == len(reg_methods)
    for i, func in enumerate(reg_methods):
        if func in ('center_of_mass', center_of_mass):
            # can't return metric
            with pytest.raises(ValueError, match='cannot return any quality'):
                affine_registration(
                    moving, static, moving_affine, static_affine,
                    pipeline=[func], ret_metric=True)
            continue

        xformed, affine_mat, \
            xopt, fopt = affine_registration(moving, static,
                                             moving_affine=moving_affine,
                                             static_affine=static_affine,
                                             level_iters=[5, 5],
                                             sigmas=[3, 1],
                                             factors=[2, 1],
                                             pipeline=[func],
                                             ret_metric=True)
        # Expected number of optimization parameters
        npt.assert_equal(len(xopt), expected_nparams[i])
        # Optimization metric must be a single numeric value
        npt.assert_equal(isinstance(fopt, (int, float)), True)

    with pytest.raises(ValueError):
        # For array input, must provide affines:
        xformed, affine_mat = affine_registration(moving, static)

    # Not supported transform names should raise an error
    npt.assert_raises(ValueError, affine_registration, moving, static,
                      moving_affine, static_affine,
                      pipeline=["wrong_transform"])

    # If providing nifti image objects, don't need to provide affines:
    moving_img = nib.Nifti1Image(moving, moving_affine)
    static_img = nib.Nifti1Image(static, static_affine)
    xformed, affine_mat = affine_registration(moving_img, static_img)
    npt.assert_almost_equal(affine_mat[:3, :3], np.eye(3), decimal=1)

    # Using strings with full paths as inputs also works:
    t1_name, b0_name = dpd.get_fnames('syn_data')
    moving = b0_name
    static = t1_name
    xformed, affine_mat = affine_registration(moving, static,
                                              level_iters=[5, 5],
                                              sigmas=[3, 1],
                                              factors=[4, 2])
    npt.assert_almost_equal(affine_mat[:3, :3], np.eye(3), decimal=1)


def test_single_transforms():
    moving = subset_b0
    static = subset_b0
    moving_affine = static_affine = np.eye(4)

    reg_methods = [center_of_mass, translation, rigid_isoscaling,
                   rigid_scaling, rigid, affine]

    for func in reg_methods:
        xformed, affine_mat = func(moving, static, moving_affine,
                                   static_affine, level_iters=[5, 5],
                                   sigmas=[3, 1], factors=[2, 1])
        # We don't ask for much:
        npt.assert_almost_equal(affine_mat[:3, :3], np.eye(3), decimal=1)


def test_register_series():
    fdata, fbval, fbvec = dpd.get_fnames('small_64D')
    img = nib.load(fdata)
    gtab = dpg.gradient_table(fbval, fbvec)
    ref_idx = np.where(gtab.b0s_mask)[0][0]
    xformed, affines = register_series(img, ref_idx)
    npt.assert_(np.all(affines[..., ref_idx] == np.eye(4)))
    npt.assert_(np.all(xformed[..., ref_idx] == img.get_fdata()[..., ref_idx]))


def test_register_dwi_series_and_motion_correction():
    fdata, fbval, fbvec = dpd.get_fnames('small_64D')
    with TemporaryDirectory() as tmpdir:
        # Use an abbreviated data-set:
        img = nib.load(fdata)
        data = img.get_fdata()[..., :10]
        nib.save(nib.Nifti1Image(data, img.affine),
                 op.join(tmpdir, 'data.nii.gz'))
        # Save a subset:
        bvals = np.loadtxt(fbval)
        bvecs = np.loadtxt(fbvec)
        np.savetxt(op.join(tmpdir, 'bvals.txt'), bvals[:10])
        np.savetxt(op.join(tmpdir, 'bvecs.txt'), bvecs[:10])
        gtab = dpg.gradient_table(op.join(tmpdir, 'bvals.txt'),
                                  op.join(tmpdir, 'bvecs.txt'))
        reg_img, reg_affines = register_dwi_series(data, gtab, img.affine)
        reg_img_2, reg_affines_2 = motion_correction(data, gtab, img.affine)
        npt.assert_(isinstance(reg_img, nib.Nifti1Image))

        npt.assert_array_equal(reg_img.get_fdata(), reg_img_2.get_fdata())
        npt.assert_array_equal(reg_affines, reg_affines_2)


@set_random_number_generator()
def test_streamline_registration(rng):
    sl1 = [np.array([[0, 0, 0], [0, 0, 0.5], [0, 0, 1], [0, 0, 1.5]]),
           np.array([[0, 0, 0], [0, 0.5, 0.5], [0, 1, 1]])]
    affine_mat = np.eye(4)
    affine_mat[:3, 3] = rng.standard_normal(3)
    sl2 = list(transform_tracking_output(sl1, affine_mat))
    aligned, matrix = streamline_registration(sl2, sl1)
    npt.assert_almost_equal(matrix, np.linalg.inv(affine_mat))
    npt.assert_almost_equal(aligned[0], sl1[0])
    npt.assert_almost_equal(aligned[1], sl1[1])

    # We assume the two tracks come from the same space, but it might have
    # some affine associated with it:
    base_aff = np.eye(4) * rng.random()
    base_aff[:3, 3] = np.array([1, 2, 3])
    base_aff[3, 3] = 1

    with TemporaryDirectory() as tmpdir:
        for use_aff in [None, base_aff]:
            fname1 = op.join(tmpdir, 'sl1.trk')
            fname2 = op.join(tmpdir, 'sl2.trk')
            if use_aff is not None:
                img = nib.Nifti1Image(np.zeros((2, 2, 2)), use_aff)
                # Move the streamlines to this other space, and report it:
                tgm1 = StatefulTractogram(
                    transform_tracking_output(sl1, np.linalg.inv(use_aff)),
                    img,
                    Space.VOX)

                save_trk(tgm1, fname1, bbox_valid_check=False)

                tgm2 = StatefulTractogram(
                    transform_tracking_output(sl2, np.linalg.inv(use_aff)),
                    img,
                    Space.VOX)

                save_trk(tgm2, fname2, bbox_valid_check=False)

            else:
                img = nib.Nifti1Image(np.zeros((2, 2, 2)), np.eye(4))
                tgm1 = StatefulTractogram(sl1, img, Space.RASMM)
                tgm2 = StatefulTractogram(sl2, img, Space.RASMM)
                save_trk(tgm1, fname1, bbox_valid_check=False)
                save_trk(tgm2, fname2, bbox_valid_check=False)

            aligned, matrix = streamline_registration(fname2, fname1)
            npt.assert_almost_equal(aligned[0], sl1[0], decimal=5)
            npt.assert_almost_equal(aligned[1], sl1[1], decimal=5)


def test_register_dwi_series_multi_b0():
    # Test if register_dwi_series works with multiple b0 images
    dwi_fname, dwi_bval_fname, \
        dwi_bvec_fname = dpd.get_fnames('sherbrooke_3shell')
    data, affine = load_nifti(dwi_fname)
    bvals, bvecs = read_bvals_bvecs(dwi_bval_fname, dwi_bvec_fname)

    data_small = data[..., :3]
    data_small = np.concatenate([data[..., :1], data_small], axis=-1)
    bvals_small = np.concatenate([bvals[:1], bvals[:3]], axis=0)
    bvecs_small = np.concatenate([bvecs[:1], bvecs[:3]], axis=0)
    gtab = dpg.gradient_table(bvals_small, bvecs_small)
    _ = motion_correction(data_small, gtab, affine)



__all__ = ["MinimumAverageDirectFlipMetric", "Metric", "CosineMetric",
           "AveragePointwiseEuclideanMetric", "EuclideanMetric", "dist",
            "mdf", "mean_manhattan_distance", "mean_euclidean_distance"]

import numpy as np

from dipy.segment.metricspeed import (SumPointwiseEuclideanMetric,
                                      MinimumAverageDirectFlipMetric,
                                      AveragePointwiseEuclideanMetric,
                                      CosineMetric, Metric,
                                      dist)

# Creates aliases
EuclideanMetric = SumPointwiseEuclideanMetric


def mdf(s1, s2):
    """ Computes the MDF (Minimum average Direct-Flip) distance
    [Garyfallidis12]_ between two streamlines.

    Streamlines must have the same number of points.

    Parameters
    ----------
    s1 : 2D array
        A streamline (sequence of N-dimensional points).
    s2 : 2D array
        A streamline (sequence of N-dimensional points).

    Returns
    -------
    double
        Distance between two streamlines.

    References
    ----------
    .. [Garyfallidis12] Garyfallidis E. et al., QuickBundles a method for
                        tractography simplification, Frontiers in Neuroscience,
                        vol 6, no 175, 2012.
    """
    return dist(MinimumAverageDirectFlipMetric(), s1, s2)


def mean_manhattan_distance(a, b):
    """ Compute the average Manhattan-L1 distance (MDF without flip)

    Arrays are representing a single streamline or a list of streamlines
    that have the same number of N-dimensional points (two last axis).

    Parameters
    ----------
    a : 2D or 3D array
        A streamline or concatenated streamlines
        (array of S streamlines by P points in N dimension).
    b : 2D or 3D array
        A streamline or concatenated streamlines
        (array of S streamlines by P points in N dimension).

    Returns
    -------
    1D array
        Distance between each S streamlines
    """
    return np.mean(np.sum(np.abs(a - b), axis=-1), axis=-1)


def mean_euclidean_distance(a, b):
    """ Compute the average Euclidean-L2 distance (MDF without flip)

    Arrays are representing a single streamline or a list of streamlines
    that have the same number of N-dimensional points (two last axis).

    Parameters
    ----------
    a : 2D or 3D array
        A streamline or concatenated streamlines
        (array of S streamlines by P points in N dimension).
    b : 2D or 3D array
        A streamline or concatenated streamlines
        (array of S streamlines by P points in N dimension).

    Returns
    -------
    1D array
        Distance between each S streamlines
    """
    return np.mean(np.sqrt(np.sum(np.square(a - b), axis=-1)), axis=-1)



import tempfile
import os
from numpy.testing import assert_raises, assert_equal
import pytest

from dipy.align.streamwarp import bundlewarp, bundlewarp_vector_filed
from dipy.data import read_five_af_bundles, two_cingulum_bundles
from dipy.tracking.streamline import (set_number_of_points, unlist_streamlines,
                                      Streamlines)
from dipy.utils.optpkg import optional_package


_, have_matplotlib, _ = optional_package("matplotlib")
fury, have_fury, _ = optional_package('fury', min_version="0.9.0")
if have_fury:
    from dipy.viz.streamline import (show_bundles, viz_two_bundles,
                                     viz_displacement_mag, viz_vector_field)
    from dipy.viz import window

bundles = read_five_af_bundles()


@pytest.mark.skipif(not have_fury or not have_matplotlib,
                    reason='Requires FURY and Matplotlib')
def test_output_created():
    views = ['axial', 'sagital', 'coronal']

    colors = [[0.91, 0.26, 0.35], [0.99, 0.50, 0.38], [0.99, 0.88, 0.57],
              [0.69, 0.85, 0.64], [0.51, 0.51, 0.63]]

    with tempfile.TemporaryDirectory() as temp_dir:
        for view in views:
            fname = os.path.join(temp_dir, f'test_{view}.png')
            show_bundles(bundles, False, view=view, save_as=fname)
            assert_equal(os.path.exists(fname), True)

        fname = os.path.join(temp_dir, 'test_colors.png')
        show_bundles(bundles, False, colors=colors, save_as=fname)
        assert_equal(os.path.exists(fname), True)

        # Check rendered image is not empty
        report = window.analyze_snapshot(fname, find_objects=True)
        assert_equal(report.objects > 0, True)

        cb1, cb2 = two_cingulum_bundles()

        fname = os.path.join(temp_dir, 'test_two_bundles.png')
        viz_two_bundles(cb1, cb2, fname=fname)
        assert_equal(os.path.exists(fname), True)


@pytest.mark.skipif(not have_fury, reason='Requires FURY')
def test_incorrect_view():
    assert_raises(ValueError, show_bundles, bundles, False, 'wrong_view')


@pytest.mark.skipif(not have_fury or not have_matplotlib,
                    reason='Requires FURY, and Matplotlib')
def test_bundlewarp_viz():

    with tempfile.TemporaryDirectory() as temp_dir:

        cingulum_bundles = two_cingulum_bundles()

        cb1 = cingulum_bundles[0]
        cb1 = Streamlines(set_number_of_points(cb1, 20))

        cb2 = cingulum_bundles[1]
        cb2 = Streamlines(set_number_of_points(cb2, 20))

        deformed_bundle, affine_bundle, _, _, _ = bundlewarp(cb1, cb2)

        offsets, directions, colors = bundlewarp_vector_filed(affine_bundle,
                                                              deformed_bundle)
        points_aligned, _ = unlist_streamlines(affine_bundle)

        fname = os.path.join(temp_dir, 'test_vector_field.png')
        viz_vector_field(points_aligned, directions, colors, offsets, fname)
        assert_equal(os.path.exists(fname), True)

        fname = os.path.join(temp_dir, 'test_mag_viz.png')
        viz_displacement_mag(affine_bundle, offsets, fname)
        assert_equal(os.path.exists(fname), True)



from dipy.utils.optpkg import optional_package
from dipy.viz.horizon.tab import (HorizonTab, build_label, color_single_slider)

fury, has_fury, setup_module = optional_package('fury', min_version="0.9.0")

if has_fury:
    from fury import ui


class ROIsTab(HorizonTab):
    def __init__(self, contour_actors):
        self.__actors = contour_actors
        self.__name = 'ROIs'

        self.__tab_id = 0
        self.__tab_ui = None

        self.__slider_label_opacity = build_label(text='Opacity')

        opacity = 1

        length = 450
        lw = 3
        radius = 8
        fs = 16
        tt = '{ratio:.0%}'

        self.__slider_opacity = ui.LineSlider2D(
            initial_value=opacity, max_value=1., length=length, line_width=lw,
            outer_radius=radius, font_size=fs, text_template=tt)

        color_single_slider(self.__slider_opacity)

        self.__slider_opacity.on_change = self.__change_opacity

    def __change_opacity(self, slider):
        opacity = slider.value
        for contour in self.__actors:
            contour.GetProperty().SetOpacity(opacity)

    def build(self, tab_id, tab_ui):
        self.__tab_id = tab_id
        self.__tab_ui = tab_ui

        self.__tab_ui.add_element(
            self.__tab_id, self.__slider_label_opacity, (.02, .85))
        self.__tab_ui.add_element(
            self.__tab_id, self.__slider_opacity, (.12, .85))

    @property
    def name(self):
        return self.__name



#!/usr/bin/python
"""
Class and helper functions for fitting the EVAC+ model.
"""
import logging
import numpy as np

from dipy.data import get_fnames
from dipy.testing.decorators import doctest_skip_parser
from dipy.utils.optpkg import optional_package
from dipy.align.reslice import reslice
from dipy.nn.utils import normalize, set_logger_level, transform_img
from dipy.nn.utils import recover_img, correct_minor_errors

tf, have_tf, _ = optional_package('tensorflow', min_version='2.0.0')
if have_tf:
    from tensorflow.keras.models import Model
    from tensorflow.keras.layers import Softmax, Conv3DTranspose
    from tensorflow.keras.layers import Conv3D, LayerNormalization, ReLU
    from tensorflow.keras.layers import Concatenate, Layer, Dropout, Add
else:
    class Model:
        pass

    class Layer:
        pass
    logging.warning('This model requires Tensorflow.\
                    Please install these packages using \
                    pip. If using mac, please refer to this \
                    link for installation. \
                    https://github.com/apple/tensorflow_macos')

logging.basicConfig()
logger = logging.getLogger('EVAC+')

def prepare_img(image):
    r"""
    Function to prepare image for model input
    Specific to EVAC+

    Parameters
    ----------
    image : np.ndarray
        Input image

    Returns
    -------
    input_data : dict
    """
    input1 = np.moveaxis(image, -1, 0)
    input1 = np.expand_dims(input1, -1)

    input2, _ = reslice(image, np.eye(4),
                        (1, 1, 1), (2, 2, 2))
    input2 = np.moveaxis(input2, -1, 0)
    input2 = np.expand_dims(input2, -1)

    input3, _ = reslice(image, np.eye(4),
                        (1, 1, 1), (4, 4, 4))
    input3 = np.moveaxis(input3, -1, 0)
    input3 = np.expand_dims(input3, -1)

    input4, _ = reslice(image, np.eye(4),
                        (1, 1, 1), (8, 8, 8))
    input4 = np.moveaxis(input4, -1, 0)
    input4 = np.expand_dims(input4, -1)

    input5, _ = reslice(image, np.eye(4),
                        (1, 1, 1), (16, 16, 16))
    input5 = np.moveaxis(input5, -1, 0)
    input5 = np.expand_dims(input5, -1)

    input_data = {"input_1":input1,
                  "input_2":input2,
                  "input_3":input3,
                  "input_4":input4,
                  "input_5":input5}

    return input_data

class Block(Layer):
    def __init__(self, out_channels, kernel_size, strides,
                 padding, drop_r, n_layers, layer_type='down'):
        super(Block, self).__init__()
        self.layer_list = []
        self.layer_list2 = []
        self.n_layers = n_layers
        for _ in range(n_layers):
            self.layer_list.append(Conv3D(out_channels,
                                          kernel_size,
                                          strides=strides,
                                          padding=padding))
            self.layer_list.append(Dropout(drop_r))
            self.layer_list.append(LayerNormalization())
            self.layer_list.append(ReLU())
        if layer_type=='down':
            self.layer_list2.append(Conv3D(1, 2, strides=2, padding='same'))
            self.layer_list2.append(ReLU())
        elif layer_type=='up':
            self.layer_list2.append(Conv3DTranspose(1, 2, strides=2, padding='same'))
            self.layer_list2.append(ReLU())

        self.channel_sum = ChannelSum()
        self.add = Add()

    def call(self, input, passed):
        x = input
        for layer in self.layer_list:
            x = layer(x)

        x = self.channel_sum(x)
        fwd = self.add([x, passed])
        x = fwd

        for layer in self.layer_list2:
            x = layer(x)

        return fwd, x

class ChannelSum(Layer):
    def __init__(self):
        super(ChannelSum, self).__init__()

    def call(self, inputs):
        return tf.reduce_sum(inputs, axis=-1, keepdims=True)

def init_model(model_scale=16):
    r"""
    Function to create model for EVAC+

    Parameters
    ----------
    model_scale : int, optional
        The scale of the model
        Should match the saved weights from fetcher
        Default is 16

    Returns
    -------
    model : tf.keras.Model
    """
    inputs = tf.keras.Input(shape=(128, 128, 128, 1))
    raw_input_2 = tf.keras.Input(shape=(64, 64, 64, 1))
    raw_input_3 = tf.keras.Input(shape=(32, 32, 32, 1))
    raw_input_4 = tf.keras.Input(shape=(16, 16, 16, 1))
    raw_input_5 = tf.keras.Input(shape=(8, 8, 8, 1))
    # Encode
    fwd1, x = Block(model_scale, kernel_size=5,
                   strides=1, padding='same',
                   drop_r=0.2, n_layers=1)(inputs, inputs)

    x = Concatenate()([x, raw_input_2])

    fwd2, x = Block(model_scale*2, kernel_size=5,
                    strides=1, padding='same',
                    drop_r=0.5, n_layers=2)(x, x)

    x = Concatenate()([x, raw_input_3])

    fwd3, x = Block(model_scale*4, kernel_size=5,
                    strides=1, padding='same',
                    drop_r=0.5, n_layers=3)(x, x)

    x = Concatenate()([x, raw_input_4])

    fwd4, x = Block(model_scale*8, kernel_size=5,
                    strides=1, padding='same',
                    drop_r=0.5, n_layers=3)(x, x)

    x = Concatenate()([x, raw_input_5])

    _, up = Block(model_scale*16, kernel_size=5,
                  strides=1, padding='same',
                  drop_r=0.5, n_layers=3,
                  layer_type='up')(x, x)

    x = Concatenate()([fwd4, up])

    _, up = Block(model_scale*8, kernel_size=5,
                  strides=1, padding='same',
                  drop_r=0.5, n_layers=3,
                  layer_type='up')(x, up)

    x = Concatenate()([fwd3, up])

    _, up = Block(model_scale*4, kernel_size=5,
                  strides=1, padding='same',
                  drop_r=0.5, n_layers=3,
                  layer_type='up')(x, up)

    x = Concatenate()([fwd2, up])

    _, up = Block(model_scale*2, kernel_size=5,
                  strides=1, padding='same',
                  drop_r=0.5, n_layers=2,
                  layer_type='up')(x, up)

    x = Concatenate()([fwd1, up])

    _, pred = Block(model_scale, kernel_size=5,
                    strides=1, padding='same',
                    drop_r=0.5, n_layers=1,
                    layer_type='none')(x, up)

    pred = Conv3D(2, 1, padding='same')(pred)
    output = Softmax(axis=-1)(pred)

    model = Model({"input_1":inputs,
                   "input_2":raw_input_2,
                   "input_3":raw_input_3,
                   "input_4":raw_input_4,
                   "input_5":raw_input_5},
                  output[..., 0])
    return model


class EVACPlus:
    """
    This class is intended for the EVAC+ model.
    """

    @doctest_skip_parser
    def __init__(self, verbose=False):
        r"""
        The model was pre-trained for usage on
        brain extraction of T1 images.

        This model is designed to take as input
        a T1 weighted image.

        Parameters
        ----------
        verbose : bool (optional)
            Whether to show information about the processing.
            Default: False

        References
        ----------
        ..  [1] Park, J.S., Fadnavis, S., & Garyfallidis, E. (2022).
                EVAC+: Multi-scale V-net with Deep Feature
                CRF Layers for Brain Extraction.
        """

        if not have_tf:
            raise tf()

        log_level = 'INFO' if verbose else 'CRITICAL'
        set_logger_level(log_level, logger)

        # EVAC+ network load

        self.model = init_model()
        self.fetch_default_weights()

    def fetch_default_weights(self):
        r"""
        Load the model pre-training weights to use for the fitting.
        While the user can load different weights, the function
        is mainly intended for the class function 'predict'.
        """
        fetch_model_weights_path = get_fnames('evac_default_weights')
        print('fetched ' + fetch_model_weights_path)
        self.load_model_weights(fetch_model_weights_path)

    def load_model_weights(self, weights_path):
        r"""
        Load the custom pre-training weights to use for the fitting.

        Parameters
        ----------
        weights_path : str
            Path to the file containing the weights (hdf5, saved by tensorflow)
        """
        try:
            self.model.load_weights(weights_path)
        except ValueError:
            raise ValueError('Expected input for the provided model weights \
                             do not match the declared model')

    def __predict(self, x_test):
        r"""
        Internal prediction function

        Parameters
        ----------
        x_test : np.ndarray (batch, 128, 128, 128, 1)
            Image should match the required shape of the model.

        Returns
        -------
        np.ndarray (batch, ...)
            Predicted brain mask
        """

        return self.model.predict(x_test)

    def predict(self, T1, affine,
                voxsize=(1, 1, 1), batch_size=None,
                return_affine=False, return_prob=False,
                largest_area=True):
        r"""
        Wrapper function to facilitate prediction of larger dataset.

        Parameters
        ----------
        T1 : np.ndarray or list of np.ndarrys
            For a single image, input should be a 3D array.
            If multiple images, it should be a a list or tuple.

        affine : np.ndarray (4, 4) or (batch, 4, 4)
            or list of np.ndarrays with len of batch
            Affine matrix for the T1 image. Should have
            batch dimension if T1 has one.

        voxsize : np.ndarray or list or tuple, optional
            (3,) or (batch, 3)
            voxel size of the T1 image.
            Default is (1, 1, 1)

        batch_size : int, optional
            Number of images per prediction pass. Only available if data
            is provided with a batch dimension.
            Consider lowering it if you get an out of memory error.
            Increase it if you want it to be faster and have a lot of data.
            If None, batch_size will be set to 1 if the provided image
            has a batch dimension.
            Default is None

        return_affine : bool, optional
            Whether to return the affine matrix. Useful if the input was a
            file path.
            Default is False

        return_prob : bool, optional
            Whether to return the probability map instead of a
            binary mask. Useful for testing.
            Default is False

        largest_area : bool, optional
            Whether to exclude only the largest background/foreground.
            Useful for solving minor errors.
            Default is True

        Returns
        -------
        pred_output : np.ndarray (...) or (batch, ...)
            Predicted brain mask

        affine : np.ndarray (...) or (batch, ...)
            affine matrix of mask
            only if return_affine is True

        """

        voxsize = np.array(voxsize)
        affine = np.array(affine)

        if isinstance(T1, (list, tuple)):
            dim = 4
            T1 = np.array(T1)
        elif len(T1.shape) == 3:
            dim = 3
            if batch_size is not None:
                logger.warning('Batch size specified, but not used',
                               'due to the input not having \
                                a batch dimension')
            batch_size = 1

            T1 = np.expand_dims(T1, 0)
            affine = np.expand_dims(affine, 0)
            voxsize = np.expand_dims(voxsize, 0)
        else:
            raise ValueError("T1 data should be a np.ndarray of dimension 3 "
                             "or a list/tuple of it")

        input_data = np.zeros((128, 128, 128, len(T1)))
        rev_affine = np.zeros((len(T1), 4, 4))
        ori_shapes = np.zeros((len(T1), 3)).astype(int)

        # Normalize the data.
        n_T1 = np.zeros(T1.shape)
        for i, T1_img in enumerate(T1):
            n_T1[i] = normalize(T1_img, new_min=0, new_max=1)
            t_img, t_affine, ori_shape = transform_img(n_T1[i],
                                                       affine[i],
                                                       voxsize[i])
            input_data[..., i] = t_img
            rev_affine[i] = t_affine
            ori_shapes[i] = ori_shape

        # Prediction stage
        prediction = np.zeros((len(T1), 128, 128, 128),
                              dtype=np.float32)
        for batch_idx in range(batch_size, len(T1)+1, batch_size):
            batch = input_data[..., batch_idx-batch_size:batch_idx]
            temp_input = prepare_img(batch)
            temp_pred = self.__predict(temp_input)
            prediction[:batch_idx] = temp_pred
        remainder = np.mod(len(T1), batch_size)
        if remainder != 0:
            temp_input = prepare_img(input_data[..., -remainder:])
            temp_pred = self.__predict(temp_input)
            prediction[-remainder:] = temp_pred

        output_mask = []
        for i, T1_img in enumerate(T1):
            output = recover_img(prediction[i],
                                 rev_affine[i],
                                 voxsize=voxsize[i],
                                 ori_shape=ori_shapes[i],
                                 image_shape=T1_img.shape)
            if not return_prob:
                output = np.where(output >= 0.5, 1, 0)
                if largest_area:
                    output = correct_minor_errors(output)
            output_mask.append(output)

        if dim == 3:
            output_mask = output_mask[0]
            affine = affine[0]

        output_mask = np.array(output_mask)
        affine = np.array(affine)
        if return_affine:
            return output_mask, affine
        else:
            return output_mask



import warnings

import numpy as np
import numpy.testing as npt

from dipy.direction.peaks import default_sphere, peaks_from_model
from dipy.reconst.shm import descoteaux07_legacy_msg
from dipy.testing.decorators import set_random_number_generator


@set_random_number_generator()
def test_EuDXDirectionGetter(rng):
    class SillyModel:
        def fit(self, data, mask=None):
            return SillyFit(self)

    class SillyFit:

        def __init__(self, model):
            self.model = model

        def odf(self, sphere):
            odf = np.zeros(sphere.theta.shape)
            r = rng.integers(0, len(odf))
            odf[r] = 1
            return odf

    def get_direction(dg, point, direction):
        newdir = direction.copy()
        state = dg.get_direction(point, newdir)
        return state, np.array(newdir)

    data = rng.random((3, 4, 5, 2))
    with warnings.catch_warnings():
        warnings.filterwarnings(
            "ignore", message=descoteaux07_legacy_msg,
            category=PendingDeprecationWarning)
        peaks = peaks_from_model(SillyModel(), data, default_sphere,
                                 relative_peak_threshold=.5,
                                 min_separation_angle=25)
    peaks._initialize()

    up = np.zeros(3)
    up[2] = 1.
    down = -up

    for i in range(3 - 1):
        for j in range(4 - 1):
            for k in range(5 - 1):
                point = np.array([i, j, k], dtype=float)

                # Test that the angle threshold rejects points
                peaks.ang_thr = 0.
                state, nd = get_direction(peaks, point, up)
                npt.assert_equal(state, 1)

                # Here we leverage the fact that we know Hemispheres project
                # all their vertices into the z >= 0 half of the sphere.
                peaks.ang_thr = 90.
                state, nd = get_direction(peaks, point, up)
                npt.assert_equal(state, 0)
                expected_dir = peaks.peak_dirs[i, j, k, 0]
                npt.assert_array_almost_equal(nd, expected_dir)
                state, nd = get_direction(peaks, point, down)
                npt.assert_array_almost_equal(nd, -expected_dir)

                # Check that we can get directions at non-integer points
                point += rng.random(3)
                state, nd = get_direction(peaks, point, up)
                npt.assert_equal(state, 0)

                # Check that points are rounded to get initial direction
                point -= .5
                initial_dir = peaks.initial_direction(point)
                # It should be a (1, 3) array
                npt.assert_array_almost_equal(initial_dir, [expected_dir])

    with warnings.catch_warnings():
        warnings.filterwarnings(
            "ignore", message=descoteaux07_legacy_msg,
            category=PendingDeprecationWarning)
        peaks1 = peaks_from_model(SillyModel(), data, default_sphere,
                                  relative_peak_threshold=.5,
                                  min_separation_angle=25,
                                  npeaks=1)
    peaks1._initialize()
    point = np.array([1, 1, 1], dtype=float)

    # it should have one direction
    npt.assert_array_almost_equal(len(peaks1.initial_direction(point)), 1)
    npt.assert_array_almost_equal(len(peaks.initial_direction(point)), 1)



"""
============================================================================
Tracking with Robust Unbiased Model-BAsed Spherical Deconvolution (RUMBA-SD)
============================================================================

Here, we demonstrate fiber tracking using a probabilistic direction getter
and RUMBA-SD, a model introduced in [CanalesRodriguez2015]_. This model adapts
Richardson-Lucy deconvolution by assuming Rician or Noncentral Chi noise
instead of Gaussian, which more accurately reflects the noise from MRI
scanners (see also
:ref:`sphx_glr_examples_built_reconstruction_reconst_rumba.py`). This tracking
tutorial is an extension on
:ref:`sphx_glr_examples_built_fiber_tracking_tracking_probabilistic.py`.

We start by loading sample data and identifying a fiber response function.
"""

from numpy.linalg import inv
import matplotlib.pyplot as plt

from dipy.core.gradients import gradient_table
from dipy.data import get_fnames, small_sphere
from dipy.direction import ProbabilisticDirectionGetter
from dipy.io.gradients import read_bvals_bvecs
from dipy.io.image import load_nifti, load_nifti_data
from dipy.io.stateful_tractogram import Space, StatefulTractogram
from dipy.io.streamline import save_trk
from dipy.reconst.csdeconv import auto_response_ssst
from dipy.tracking import utils
from dipy.tracking.local_tracking import LocalTracking
from dipy.tracking.streamline import Streamlines, transform_streamlines
from dipy.tracking.stopping_criterion import ThresholdStoppingCriterion
from dipy.viz import window, actor, colormap
from dipy.reconst.rumba import RumbaSDModel


# Enables/disables interactive visualization
interactive = False

hardi_fname, hardi_bval_fname, hardi_bvec_fname = get_fnames('stanford_hardi')
label_fname = get_fnames('stanford_labels')
t1_fname = get_fnames('stanford_t1')

data, affine, hardi_img = load_nifti(hardi_fname, return_img=True)
labels = load_nifti_data(label_fname)
t1_data, t1_aff = load_nifti(t1_fname)
bvals, bvecs = read_bvals_bvecs(hardi_bval_fname, hardi_bvec_fname)
gtab = gradient_table(bvals, bvecs)

seed_mask = (labels == 2)
white_matter = (labels == 1) | (labels == 2)
seeds = utils.seeds_from_mask(seed_mask, affine, density=2)

response, ratio = auto_response_ssst(gtab, data, roi_radii=10, fa_thr=0.7)

sphere = small_sphere

###############################################################################
# We can now initialize a `RumbaSdModel` model and fit it globally by setting
# `voxelwise` to `False`. For this example, TV regularization (`use_tv`) will
# be turned off for efficiency, although its usage can provide more coherent
# results in fiber tracking. The fit will take about 5 minutes to complete.

rumba = RumbaSDModel(gtab, wm_response=response[0], n_iter=200,
                     voxelwise=False, use_tv=False, sphere=sphere)
rumba_fit = rumba.fit(data, mask=white_matter)
odf = rumba_fit.odf()  # fODF
f_wm = rumba_fit.f_wm  # white matter volume fractions

###############################################################################
# To establish stopping criterion, a common technique is to use the Generalized
# Fractional Anisotropy (GFA). One point of caution is that RUMBA-SD by default
# separates the fODF from an isotropic compartment. This can bias GFA results
# computed on the fODF, although it will still generally be an effective
# criterion.
#
# However, an alternative stopping criterion that takes advantage of this
# feature is to use RUMBA-SD's white matter volume fraction map.

stopping_criterion = ThresholdStoppingCriterion(f_wm, .25)

###############################################################################
# We can visualize a slice of this mask.

sli = f_wm.shape[2] // 2
plt.figure()

plt.subplot(1, 2, 1).set_axis_off()
plt.imshow(f_wm[:, :, sli].T, cmap='gray', origin='lower')

plt.subplot(1, 2, 2).set_axis_off()
plt.imshow((f_wm[:, :, sli] > 0.25).T, cmap='gray', origin='lower')

plt.savefig('f_wm_tracking_mask.png')

###############################################################################
# .. rst-class:: centered small fst-italic fw-semibold
#
# White matter volume fraction slice
#
#
#
# These discrete fODFs can be used as a PMF in the
# `ProbabilisticDirectionGetter` for sampling tracking directions. The PMF
# must be strictly non-negative; RUMBA-SD already adheres to this constraint
# so no further manipulation of the fODFs is necessary.

prob_dg = ProbabilisticDirectionGetter.from_pmf(odf, max_angle=30.,
                                                sphere=sphere)
streamline_generator = LocalTracking(prob_dg, stopping_criterion, seeds,
                                     affine, step_size=.5)
streamlines = Streamlines(streamline_generator)

color = colormap.line_colors(streamlines)
streamlines_actor = actor.streamtube(
    list(transform_streamlines(streamlines, inv(t1_aff))),
    color, linewidth=0.1)

vol_actor = actor.slicer(t1_data)
vol_actor.display(x=40)
vol_actor2 = vol_actor.copy()
vol_actor2.display(z=35)

scene = window.Scene()
scene.add(vol_actor)
scene.add(vol_actor2)
scene.add(streamlines_actor)
if interactive:
    window.show(scene)

window.record(scene, out_path='tractogram_probabilistic_rumba.png',
              size=(800, 800))

sft = StatefulTractogram(streamlines, hardi_img, Space.RASMM)
save_trk(sft, "tractogram_probabilistic_rumba.trk")

###############################################################################
# .. rst-class:: centered small fst-italic fw-semibold
#
# RUMBA-SD tractogram
#
#
#
# References
# ----------
#
# .. [CanalesRodriguez2015] Canales-Rodríguez, E. J., Daducci, A.,
#    Sotiropoulos, S. N., Caruyer, E., Aja-Fernández, S., Radua, J.,
#    Mendizabal, J. M. Y., Iturria-Medina, Y., Melie-García, L., Alemán-Gómez,
#    Y., Thiran, J.-P., Sarró, S., Pomarol-Clotet, E., & Salvador, R. (2015).
#    Spherical Deconvolution of Multichannel Diffusion MRI Data with
#    Non-Gaussian Noise Models and Spatial Regularization. PLOS ONE, 10(10),
#    e0138910. https://doi.org/10.1371/journal.pone.0138910



"""
==============================================
Denoise images using Non-Local Means (NLMEANS)
==============================================

Using the non-local means filter [Coupe08]_ and [Coupe11]_ and  you can denoise
3D or 4D images and boost the SNR of your datasets. You can also decide between
modeling the noise as Gaussian or Rician (default).

We start by loading the necessary modules
"""

import numpy as np
import matplotlib.pyplot as plt
from time import time
from dipy.denoise.nlmeans import nlmeans
from dipy.denoise.noise_estimate import estimate_sigma
from dipy.data import get_fnames
from dipy.io.image import load_nifti, save_nifti

###############################################################################
# Then, let's fetch and load a T1 data from Stanford University

t1_fname = get_fnames('stanford_t1')
data, affine = load_nifti(t1_fname)

mask = data > 1500

print("vol size", data.shape)

###############################################################################
# In order to call ``non_local_means`` first you need to estimate the standard
# deviation of the noise. We have used N=32 since the Stanford dataset was
# acquired on a 3T GE scanner with a 32 array head coil.

sigma = estimate_sigma(data, N=32)

###############################################################################
# Calling the main function ``non_local_means``

t = time()

den = nlmeans(data, sigma=sigma, mask=mask, patch_radius=1,
              block_radius=2, rician=True)

print("total time", time() - t)

###############################################################################
# Let us plot the axial slice of the denoised output

axial_middle = data.shape[2] // 2

before = data[:, :, axial_middle].T
after = den[:, :, axial_middle].T

difference = np.abs(after.astype(np.float64) - before.astype(np.float64))

difference[~mask[:, :, axial_middle].T] = 0


fig, ax = plt.subplots(1, 3)
ax[0].imshow(before, cmap='gray', origin='lower')
ax[0].set_title('before')
ax[1].imshow(after, cmap='gray', origin='lower')
ax[1].set_title('after')
ax[2].imshow(difference, cmap='gray', origin='lower')
ax[2].set_title('difference')

plt.savefig('denoised.png', bbox_inches='tight')

###############################################################################
# .. rst-class:: centered small fst-italic fw-semibold
#
# Showing axial slice before (left) and after (right) NLMEANS denoising

save_nifti('denoised.nii.gz', den, affine)

###############################################################################
# An improved version of non-local means denoising is adaptive soft coefficient
# matching, please refer to
# :ref:`sphx_glr_examples_built_preprocessing_denoise_ascm.py` for more
# details.
#
# References
# ----------
#
# .. [Coupe08] P. Coupe, P. Yger, S. Prima, P. Hellier, C. Kervrann,
#    C. Barillot, "An Optimized Blockwise Non Local Means Denoising Filter
#    for 3D Magnetic Resonance Images", IEEE Transactions on Medical Imaging,
#    27(4):425-441, 2008
#
# .. [Coupe11] Pierrick Coupe, Jose Manjon, Montserrat Robles, Louis Collins.
#     "Adaptive Multiresolution Non-Local Means Filter for 3D MR Image
#     Denoising" IET Image Processing, Institution of Engineering and
#     Technology, 2011



from typing import Any
from dataclasses import dataclass
import warnings
from abc import ABC, abstractmethod

import numpy as np

from dipy.utils.optpkg import optional_package

fury, has_fury, setup_module = optional_package('fury', min_version="0.9.0")

if has_fury:
    from fury import ui
    from fury.data import read_viz_icons


@dataclass
class HorizonUIElement:
    """Dataclass to define properties of horizon ui elements.
    """
    visibility: bool
    selected_value: Any
    obj: Any
    position = (0, 0)


class HorizonTab(ABC):
    """Base for different tabs available in horizon.
    """
    def __init__(self):
        self._elements = []

    @abstractmethod
    def build(self, tab_id, tab_ui):
        """Build all the elements under the tab.

        Parameters
        ----------
        tab_id : int
            Id of the tab.
        tab_ui : TabUI
            FURY TabUI object for tabs panel.

        Notes
        -----
        tab_ui will removed once every all tabs adapt new build architecture.
        """

    def register_elements(self, *args):
        """Register elements for rendering.

        Parameters
        ----------
        *args : HorizonUIElement(s)
            Elements to be register for rendering.
        """
        for element in args:
            self._elements.append(element)

    @property
    @abstractmethod
    def name(self):
        """Name of the tab.
        """

    @property
    def elements(self):
        """list of underlying FURY ui elements in the tab.
        """
        return self._elements


class TabManager:
    """
    A Manager for tabs of the table panel.

    Attributes
    ----------

    tab_ui : TabUI
        Underlying FURY TabUI object.
    """
    def __init__(self, tabs, win_size, synchronize_slices=False):
        num_tabs = len(tabs)
        self._tabs = tabs
        self._synchronize_slices = synchronize_slices

        win_width, _win_height = win_size

        self._tab_size = (1280, 240)
        x_pad = np.rint((win_width - self._tab_size[0]) / 2)

        self._active_tab_id = num_tabs - 1

        self._tab_ui = ui.TabUI(
            position=(x_pad, 5), size=self._tab_size, nb_tabs=num_tabs,
            active_color=(1, 1, 1), inactive_color=(0.5, 0.5, 0.5),
            draggable=True, startup_tab_id=self._active_tab_id)

        self._tab_ui.on_change = self._tab_selected

        self.tab_changed = lambda actors: None
        slices_tabs = list(
            filter(
                lambda x: x.__class__.__name__ == 'SlicesTab', self._tabs
            )
        )
        if not self._synchronize_slices and slices_tabs:
            msg = 'Images are of different dimensions, ' \
                + 'synchronization of slices will not work'
            warnings.warn(msg)

        for tab_id, tab in enumerate(tabs):
            self._tab_ui.tabs[tab_id].title = ' ' + tab.name
            self._tab_ui.tabs[tab_id].title_font_size = 18
            tab.build(tab_id, self._tab_ui)
            if tab.__class__.__name__ == 'SlicesTab':
                tab.on_slice_change = self.synchronize_slices
                self._render_tab_elements(tab_id, tab.elements)

    def _render_tab_elements(self, tab_id, elements):
        for element in elements:
            if isinstance(element.position, list):
                for i, position in enumerate(element.position):
                    self._tab_ui.add_element(tab_id, element.obj[i], position)
            else:
                self._tab_ui.add_element(tab_id, element.obj, element.position)

    def _tab_selected(self, tab_ui):
        if self._active_tab_id == tab_ui.active_tab_idx:
            self._active_tab_id = -1
            return

        self._active_tab_id = tab_ui.active_tab_idx

        current_tab = self._tabs[self._active_tab_id]
        if current_tab.__class__.__name__ == 'SlicesTab':
            self.tab_changed(current_tab.actors)
            current_tab.on_tab_selected()

    def reposition(self, win_size):
        """
        Reposition the tabs panel.

        Parameters
        ----------
        win_size : (float, float)
            size of the horizon window.
        """
        win_width, _win_height = win_size
        x_pad = np.rint((win_width - self._tab_size[0]) / 2)
        self._tab_ui.position = (x_pad, 5)

    def synchronize_slices(self, active_tab_id, x_value, y_value, z_value):
        """
        Synchronize slicers for all the images

        Parameters
        ----------
        active_tab_id: int
            tab_id of the action performing tab
        x_value: float
            x-value of the active slicer
        y_value: float
            y-value of the active slicer
        z_value: float
            z-value of the active slicer
        """

        if not self._synchronize_slices:
            return

        slices_tabs = list(
            filter(
                lambda x: x.__class__.__name__ == 'SlicesTab'
                and not x.tab_id == active_tab_id, self._tabs
            )
        )

        for slices_tab in slices_tabs:
            slices_tab.update_slices(x_value, y_value, z_value)

    @property
    def tab_ui(self):
        """FURY TabUI object.
        """
        return self._tab_ui


def build_label(text, font_size=16, bold=False, is_horizon_label=False):
    """Simple utility function to build labels.

    Parameters
    ----------
    text : str
    font_size : int, optional
    bold : bool, optional

    Returns
    -------
    label : TextBlock2D
    """

    label = ui.TextBlock2D()
    label.message = text
    label.font_size = font_size
    label.font_family = 'Arial'
    label.justification = 'left'
    label.bold = bold
    label.italic = False
    label.shadow = False
    label.actor.GetTextProperty().SetBackgroundColor(0, 0, 0)
    label.actor.GetTextProperty().SetBackgroundOpacity(0.0)
    label.color = (0.7, 0.7, 0.7)

    if is_horizon_label:
        return HorizonUIElement(True, text, label)
    return label


def build_slider(
        initial_value,
        max_value,
        min_value=0,
        length=450,
        line_width=3,
        radius=8,
        font_size=16,
        text_template='{value:.1f} ({ratio:.0%})',
        on_moving_slider=lambda _slider: None,
        on_value_changed=lambda _slider: None,
        on_change=lambda _slider: None,
        label='',
        label_font_size=16,
        label_style_bold=False,
        is_double_slider=False
):
    """Create a horizon theme based disk-knob slider.

    Parameters
    ----------
    initial_value : float, (float, float)
        Initial value(s) of the slider.
    max_value : float
        Maximum value of the slider.
    min_value : float, optional
        Minimum value of the slider.
    length : int, optional
        Length of the slider.
    line_width : int, optional
        Width of the line on which the disk will slide.
    radius : int, optional
        Radius of the disk handle.
    font_size : int, optional
        Size of the text to display alongside the slider (pt).
    text_template : str, callable, optional
        If str, text template can contain one or multiple of the
        replacement fields: `{value:}`, `{ratio:}`.
        If callable, this instance of `:class:LineSlider2D` will be
        passed as argument to the text template function.
    on_moving_slider : callable, optional
        When the slider is interacted by the user.
    on_value_changed : callable, optional
        When value of the slider changed programmatically.
    on_change : callable, optional
        When value of the slider changed.
    label : str, optional
        Label to ui element for slider
    label_font_size : int, optional
        Size of label text to display with slider
    label_style_bold : bool, optional
        Is label should have bold style.

    Return
    ------
    (label: HorizonUIElement, element(slider): HorizonUIElement)
    """

    if is_double_slider and 'ratio' in text_template:
        warnings.warn('Double slider only support values and not ratio')
        return

    slider_label = build_label(
        label,
        font_size=label_font_size,
        bold=label_style_bold,
        is_horizon_label=True
    )

    if not is_double_slider:
        slider = ui.LineSlider2D(
            initial_value=initial_value,
            max_value=max_value,
            min_value=min_value,
            length=length,
            line_width=line_width,
            outer_radius=radius,
            font_size=font_size,
            text_template=text_template
        )
    else:
        slider = ui.LineDoubleSlider2D(
            initial_values=initial_value,
            max_value=max_value,
            min_value=min_value,
            length=length,
            line_width=line_width,
            outer_radius=radius,
            font_size=font_size,
            text_template=text_template
        )

    slider.on_moving_slider = on_moving_slider
    slider.on_value_changed = on_value_changed
    slider.on_change = on_change

    slider.default_color = (1., .5, .0)
    slider.track.color = (.8, .3, .0)
    slider.active_color = (.9, .4, .0)
    if not is_double_slider:
        slider.handle.color = (1., .5, .0)
    else:
        slider.handles[0].color = (1., .5, .0)
        slider.handles[1].color = (1., .5, .0)

    return (
        slider_label,
        HorizonUIElement(True, initial_value, slider)
    )


def build_checkbox(
        labels=None,
        checked_labels=None,
        padding=1,
        font_size=16,
        on_change=lambda _checkbox: None
):
    """Create horizon theme checkboxes.

    Parameters
    ----------
    labels : list(str), optional
        List of labels of each option.
    checked_labels: list(str), optional
        List of labels that are checked on setting up.
    padding : float, optional
        The distance between two adjacent options element
    font_size : int, optional
        Size of the text font.
    on_change : callback, optional
        When checkbox value changed

    Returns
    -------
    checkbox : HorizonUIElement
    """

    if labels is None or not labels:
        warnings.warn('At least one label needs to be to create checkboxes')
        return

    if checked_labels is None:
        checked_labels = ()

    checkboxes = ui.Checkbox(
        labels=labels,
        checked_labels=checked_labels,
        padding=padding,
        font_size=font_size
    )

    checkboxes.on_change = on_change

    return HorizonUIElement(True, checked_labels, checkboxes)


def build_switcher(
        items=None,
        label='',
        initial_selection=0,
        on_prev_clicked=lambda _selected_value: None,
        on_next_clicked=lambda _selected_value: None,
        on_value_changed=lambda _selected_idx, _selected_value: None,
):
    """Create horizon theme switcher.

    Parameters
    ----------
    items : list, optional
        dictionaries with keys 'label' and 'value'. Label will be used to show
        it to user and value will be used for selection.
    label : str, optional
        label for the switcher.
    initial_selection : int, optional
        index of the selected item initially.
    on_prev_clicked : callback, optional
        method providing a callback when prev value is selected in switcher.
    on_next_clicked : callback, optional
        method providing a callback when next value is selected in switcher.
    on_value_changed : callback, optional
        method providing a callback when either prev or next value selected in
        switcher.

    Returns
    -------
    HorizonCombineElement(
        label: HorizonUIElement,
        element(switcher): HorizonUIElement)

    Notes
    -----
    switcher: consists 'obj' which is an array providing FURY UI elements used.
    """
    if items is None:
        warnings.warn('No items passed in switcher')
        return

    num_items = len(items)

    if initial_selection >= num_items:
        initial_selection = 0

    switch_label = build_label(text=label, is_horizon_label=True)
    selection_label = build_label(
        text=items[initial_selection]['label'])

    left_button = ui.Button2D(
            icon_fnames=[('left', read_viz_icons(fname='circle-left.png'))],
            size=(25, 25))
    right_button = ui.Button2D(
            icon_fnames=[('right', read_viz_icons(fname='circle-right.png'))],
            size=(25, 25))

    switcher = HorizonUIElement(True, [initial_selection,
                                items[initial_selection]['value']],
                                [left_button, selection_label, right_button])

    def left_clicked(_i_ren, _obj, _button):
        selected_id = switcher.selected_value[0] - 1
        if selected_id < 0:
            selected_id = num_items - 1
        value_changed(selected_id)
        on_prev_clicked(items[selected_id]['value'])
        on_value_changed(selected_id, items[selected_id]['value'])

    def right_clicked(_i_ren, _obj, _button):
        selected_id = switcher.selected_value[0] + 1
        if selected_id >= num_items:
            selected_id = 0
        value_changed(selected_id)
        on_next_clicked(items[selected_id]['value'])
        on_value_changed(selected_id, items[selected_id]['value'])

    def value_changed(selected_id):
        switcher.selected_value[0] = selected_id
        switcher.selected_value[1] = items[selected_id]['value']
        selection_label.message = items[selected_id]['label']

    left_button.on_left_mouse_button_clicked = left_clicked
    right_button.on_left_mouse_button_clicked = right_clicked

    return (
        switch_label,
        switcher
    )


def color_single_slider(slider):
    slider.default_color = (1., .5, .0)
    slider.track.color = (.8, .3, .0)
    slider.active_color = (.9, .4, .0)
    slider.handle.color = (1., .5, .0)


def color_double_slider(slider):
    slider.default_color = (1., .5, .0)
    slider.track.color = (.8, .3, .0)
    slider.active_color = (.9, .4, .0)
    slider.handles[0].color = (1., .5, .0)
    slider.handles[1].color = (1., .5, .0)



from dipy.viz.horizon.tab.base import (HorizonTab, TabManager, build_label,
                                       build_slider, build_checkbox,
                                       build_switcher, color_double_slider,
                                       color_single_slider)
from dipy.viz.horizon.tab.cluster import ClustersTab
from dipy.viz.horizon.tab.peak import PeaksTab
from dipy.viz.horizon.tab.roi import ROIsTab
from dipy.viz.horizon.tab.slice import SlicesTab

__all__ = [
    'HorizonTab', 'TabManager', 'ClustersTab', 'PeaksTab', 'ROIsTab',
    'SlicesTab', 'build_label', 'build_slider', 'build_checkbox',
    'build_switcher',
    'color_double_slider', 'color_single_slider']



import pytest

from dipy.data import get_fnames
from dipy.utils.optpkg import optional_package
import numpy as np
from numpy.testing import assert_almost_equal

tf, have_tf, _ = optional_package('tensorflow', min_version='2.0.0')
tfa, have_tfa, _ = optional_package('tensorflow_addons')

if have_tf and have_tfa:
    from dipy.nn.synb0 import Synb0


@pytest.mark.skipif(not all([have_tf, have_tfa]), reason='Requires TensorFlow \
                                                          , TensorFlow_addons')
def test_default_weights():
    file_names = get_fnames('synb0_test_data')
    input_arr1 = np.load(file_names[0])['b0'][0]
    input_arr2 = np.load(file_names[0])['T1'][0]
    target_arr = np.load(file_names[1])['arr_0'][0]

    synb0_model = Synb0()
    synb0_model.fetch_default_weights(0)
    results_arr = synb0_model.predict(input_arr1, input_arr2, average=False)
    assert_almost_equal(results_arr, target_arr, decimal=1)


@pytest.mark.skipif(not all([have_tf, have_tfa]), reason='Requires TensorFlow \
                                                          , TensorFlow_addons')
def test_default_weights_batch():
    file_names = get_fnames('synb0_test_data')
    input_arr1 = np.load(file_names[0])['b0']
    input_arr2 = np.load(file_names[0])['T1']
    target_arr = np.load(file_names[1])['arr_0']
    synb0_model = Synb0()
    synb0_model.fetch_default_weights(0)
    results_arr = synb0_model.predict(input_arr1, input_arr2,
                                      batch_size=2, average=False)
    assert_almost_equal(results_arr, target_arr, decimal=1)



"""
=====================================================================
Reconstruction of the diffusion signal with the kurtosis tensor model
=====================================================================

The diffusion kurtosis model is an expansion of the diffusion tensor model
(see :ref:`sphx_glr_examples_built_reconstruction_reconst_dti.py`). In
addition to the diffusion tensor (DT), the diffusion kurtosis model
quantifies the degree to which water diffusion in biological tissues is
non-Gaussian using the kurtosis tensor (KT) [Jensen2005]_.

Measurements of non-Gaussian diffusion from the diffusion kurtosis model are of
interest because they can be used to characterize tissue microstructural
heterogeneity [Jensen2010]_. Moreover, DKI can be used to: 1) derive concrete
biophysical parameters, such as the density of axonal fibers and diffusion
tortuosity [Fierem2011]_ (see
:ref:`sphx_glr_examples_built_reconstruction_reconst_dki_micro.py`); and 2)
resolve crossing fibers in tractography and to obtain invariant rotational
measures not limited to well-aligned fiber populations [NetoHe2015]_.

The diffusion kurtosis model expresses the diffusion-weighted signal as:

.. math::

    S(n,b)=S_{0}e^{-bD(n)+\\frac{1}{6}b^{2}D(n)^{2}K(n)}

where $\mathbf{b}$ is the applied diffusion weighting (which is dependent on
the measurement parameters), $S_0$ is the signal in the absence of diffusion
gradient sensitization, $\mathbf{D(n)}$ is the value of diffusion along
direction $\mathbf{n}$, and $\mathbf{K(n)}$ is the value of kurtosis along
direction $\mathbf{n}$. The directional diffusion $\mathbf{D(n)}$ and kurtosis
$\mathbf{K(n)}$ can be related to the diffusion tensor (DT) and kurtosis tensor
(KT) using the following equations:

.. math::
     D(n)=\\sum_{i=1}^{3}\\sum_{j=1}^{3}n_{i}n_{j}D_{ij}

and

.. math::
     K(n)=\\frac{MD^{2}}{D(n)^{2}}\\sum_{i=1}^{3}\\sum_{j=1}^{3}\\sum_{k=1}^{3}
     \\sum_{l=1}^{3}n_{i}n_{j}n_{k}n_{l}W_{ijkl}

where $D_{ij}$ are the elements of the second-order DT, and $W_{ijkl}$ the
elements of the fourth-order KT and $MD$ is the mean diffusivity. As the DT,
KT has antipodal symmetry and thus only 15 Wijkl elements are needed to fully
characterize the KT:

.. math::
   \\begin{matrix} ( & W_{xxxx} & W_{yyyy} & W_{zzzz} & W_{xxxy} & W_{xxxz}
                    & ... \\
                    & W_{xyyy} & W_{yyyz} & W_{xzzz} & W_{yzzz} & W_{xxyy}
                    & ... \\
                    & W_{xxzz} & W_{yyzz} & W_{xxyz} & W_{xyyz} & W_{xyzz}
                    & & )\\end{matrix}

In the following example we show how to fit the diffusion kurtosis model on
diffusion-weighted multi-shell datasets and how to estimate diffusion kurtosis
based statistics.

First, we import all relevant modules:
"""

import numpy as np
import dipy.reconst.dki as dki
import dipy.reconst.dti as dti
from dipy.core.gradients import gradient_table
from dipy.data import get_fnames
from dipy.io.gradients import read_bvals_bvecs
from dipy.io.image import load_nifti
from dipy.segment.mask import median_otsu
from dipy.viz.plotting import compare_maps
from scipy.ndimage import gaussian_filter

###############################################################################
# DKI requires multi-shell data, i.e. data acquired from more than one
# non-zero b-value. Here, we use fetch to download a multi-shell dataset which
# was kindly provided by Hansen and Jespersen (more details about the data are
# provided in their paper [Hansen2016]_). The total size of the downloaded
# data is 192 MBytes, however you only need to fetch it once.

fraw, fbval, fbvec, t1_fname = get_fnames('cfin_multib')

data, affine = load_nifti(fraw)
bvals, bvecs = read_bvals_bvecs(fbval, fbvec)
gtab = gradient_table(bvals, bvecs)

###############################################################################
# Function ``get_fnames`` downloads and outputs the paths of the data,
# ``load_nifti`` returns the data as a nibabel Nifti1Image object, and
# ``read_bvals_bvecs`` loads the arrays containing the information about the
# b-values and b-vectors. These later arrays are converted to the
# GradientTable object required for Dipy_'s data reconstruction.
#
# Before fitting the data, we perform some data pre-processing. We first
# compute a brain mask to avoid unnecessary calculations on the background
# of the image.

maskdata, mask = median_otsu(data, vol_idx=[0, 1], median_radius=4, numpass=2,
                             autocrop=False, dilate=1)

###############################################################################
# Since the diffusion kurtosis models involves the estimation of a large
# number of parameters [TaxCMW2015]_ and since the non-Gaussian components of
# the diffusion signal are more sensitive to artefacts [NetoHe2012]_, it might
# be favorable to suppress the effects of noise and artefacts before diffusion
# kurtosis fitting. In this example the effects of noise and artefacts are
# suppress by using 3D Gaussian smoothing (with a Gaussian kernel with
# fwhm=1.25) as suggested by pioneer DKI studies (e.g. [Jensen2005]_,
# [NetoHe2012]_). Although here the Gaussian smoothing is used so that results
# are comparable to these studies, it is important to note that more advanced
# noise and artifact suppression algorithms are available in DIPY_, e.g. the
# Marcenko-Pastur PCA denoising algorithm
# (:ref:`sphx_glr_examples_built_preprocessing_denoise_mppca.py`) and
# the Gibbs artefact suppression algorithm
# (:ref:`sphx_glr_examples_built_preprocessing_denoise_gibbs.py`).

fwhm = 1.25
gauss_std = fwhm / np.sqrt(8 * np.log(2))  # converting fwhm to Gaussian std
data_smooth = np.zeros(data.shape)
for v in range(data.shape[-1]):
    data_smooth[..., v] = gaussian_filter(data[..., v], sigma=gauss_std)

###############################################################################
# Now that we have loaded and pre-processed the data we can go forward
# with DKI fitting. For this, the DKI model is first defined for the data's
# GradientTable object by instantiating the DiffusionKurtosisModel object in
# the following way:

dkimodel = dki.DiffusionKurtosisModel(gtab)

###############################################################################
# To fit the data using the defined model object, we call the ``fit`` function
# of this object. For the purpose of this example, we will only fit a
# single slice of the data:

data_smooth = data_smooth[:, :, 9:10]
mask = mask[:, :, 9:10]
dkifit = dkimodel.fit(data_smooth, mask=mask)

###############################################################################
# The fit method creates a DiffusionKurtosisFit object, which contains all the
# diffusion and kurtosis fitting parameters and other DKI attributes. For
# instance, since the diffusion kurtosis model estimates the diffusion tensor,
# all standard diffusion tensor statistics can be computed from the
# DiffusionKurtosisFit instance. For example, we can extract the fractional
# anisotropy (FA), the mean diffusivity (MD), the axial diffusivity (AD) and
# the radial diffusivity (RD) from the DiffusionKurtosisiFit instance. Of
# course, these measures can also be computed from DIPY's ``TensorModel`` fit,
# and should be analogous; however, theoretically, the diffusion statistics
# from the kurtosis model are expected to have better accuracy, since DKI's
# diffusion tensor are decoupled from higher order terms effects
# [Veraar2011]_, [NetoHe2012]_. Below we compare the FA, MD, AD, and RD,
# computed from both DTI and DKI.

tenmodel = dti.TensorModel(gtab)
tenfit = tenmodel.fit(data_smooth, mask=mask)

fits = [tenfit, dkifit]
maps = ['fa', 'md', 'ad', 'rd']
fit_labels = ['DTI', 'DKI']
map_kwargs = [{'vmax': 0.7}, {'vmax': 2e-3}, {'vmax': 2e-3}, {'vmax': 2e-3}]
compare_maps(fits, maps, fit_labels=fit_labels, map_kwargs=map_kwargs,
             filename='Diffusion_tensor_measures_from_DTI_and_DKI.png')

###############################################################################
# .. rst-class:: centered small fst-italic fw-semibold
#
# Diffusion tensor measures obtained from the diffusion tensor estimated
# from DKI (upper panels) and DTI (lower panels).
#
#
# DTI's diffusion estimates present lower values than DKI's estimates,
# showing that DTI's diffusion measurements are underestimated  by higher
# order effects.
#
# In addition to the standard diffusion statistics, the DiffusionKurtosisFit
# instance can be used to estimate the non-Gaussian measures of mean kurtosis
# (MK), the axial kurtosis (AK) and the radial kurtosis (RK).

maps = ['mk', 'ak', 'rk']
compare_maps([dkifit], maps, fit_labels=['DKI'],
             map_kwargs={'vmin': 0, 'vmax': 1.5},
             filename='Kurtosis_tensor_standard_measures.png')

###############################################################################
# .. rst-class:: centered small fst-italic fw-semibold
#
# DKI standard kurtosis measures.
#
#
# The non-Gaussian behaviour of the diffusion signal is expected to be higher
# when tissue water is confined by multiple compartments. MK is, therefore,
# higher in white matter since it is highly compartmentalized by myelin
# sheaths. These water diffusion compartmentalization is expected to be more
# pronounced perpendicularly to white matter fibers and thus the RK map
# presents higher amplitudes than the AK map.
#
# It is important to note that kurtosis estimates might present negative
# estimates in deep white matter regions (e.g. the band of dark voxels in the
# RK map above). These negative kurtosis values are artefactual and might be
# induced by:
# 1) low radial diffusivities of aligned white matter - since it is very hard
# to capture non-Gaussian information in radial direction due to it's low
# diffusion decays, radial kurtosis estimates (and consequently the mean
# kurtosis estimates) might have low robustness and tendency to exhibit
# negative values [NetoHe2012]_;
# 2) Gibbs artefacts - MRI images might be corrupted by signal oscillation
# artefact between tissue's edges if an inadequate number of high frequencies
# of the k-space is sampled. These oscillations might have different signs on
# images acquired with different diffusion-weighted and inducing negative
# biases in kurtosis parametric maps [Perron2015]_, [NetoHe2018]_.
#
# One can try to suppress this issue by using the more advance noise and
# artefact suppression algorithms, e.g., as mentioned above, the MP-PCA
# denoising (:ref:`sphx_glr_examples_built_preprocessing_denoise_mppca.py`)
# and Gibbs Unringing
# (:ref:`sphx_glr_examples_built_preprocessing_denoise_gibbs.py`)
# algorithms. Alternatively, one can overcome this artefact by computing the
# kurtosis values from powder-averaged diffusion-weighted signals. The details
# on how to compute the kurtosis from powder-average signals in dipy are
# described in follow the tutorial
# (:ref:`sphx_glr_examples_built_reconstruction_reconst_msdki.py`). Finally,
# one can use constrained optimization to ensure that the fitted parameters
# are physically plausible [DelaHa2020]_, as we will illustrate in the next
# section. Ideally though, artefacts such as Gibbs ringing should be corrected
# for as well as possible before using constrained optimization.
#
# Constrained optimization for DKI
# ================================
#
# When instantiating the DiffusionKurtosisModel, the model can be set up to use
# constraints with the option `fit_method='CLS'` (for ordinary least squares)
# or with `fit_method='CWLS'` (for weighted least squares). Constrained fitting
# takes more time than unconstrained fitting, but is generally recommended to
# prevent physically unplausible parameter estimates [DelaHa2020]_. For
# performance purposes it is recommended to use the MOSEK solver
# (https://www.mosek.com/) by setting ``cvxpy_solver='MOSEK'``. Different
# solvers can differ greatly in terms of runtime and solution accuracy, and in
# some cases solvers may show warnings about convergence or recommended option
# settings.
#
# .. note::
#    In certain atypical scenarios, the DKI+ constraints could potentially be
#    too restrictive. Always check the results of a constrained fit with their
#    unconstrained counterpart to verify that there are no unexpected
#    qualitative differences.

dkimodel_plus = dki.DiffusionKurtosisModel(gtab, fit_method='CLS')
dkifit_plus = dkimodel_plus.fit(data_smooth, mask=mask)

###############################################################################
# We can now compare the kurtosis measures obtained with the constrained fit to
# the measures obtained before, where we see that many of the artefactual
# voxels have now been corrected. In particular outliers caused by pure noise
# -- instead of for example acquisition artefacts -- can be corrected with
# this method.

compare_maps([dkifit_plus], ['mk', 'ak', 'rk'], fit_labels=['DKI+'],
             filename='Kurtosis_tensor_standard_measures_plus.png')

###############################################################################
# .. rst-class:: centered small fst-italic fw-semibold
#
# DKI standard kurtosis measures obtained with constrained optimization.
#
#
# When using constrained optimization, the expected range of the kurtosis
# measures is also naturally constrained, and so does not typically require
# additional clipping.
#
# Finally, constrained optimization obviates the need for smoothing in many
# cases:

dkifit_noisy = dkimodel.fit(data[:, :, 9:10], mask=mask)
dkifit_noisy_plus = dkimodel_plus.fit(data[:, :, 9:10], mask=mask)

compare_maps([dkifit_noisy, dkifit_noisy_plus], ['mk', 'ak', 'rk'],
             fit_labels=['DKI', 'DKI+'], map_kwargs={'vmin': 0, 'vmax': 1.5},
             filename='Kurtosis_tensor_standard_measures_noisy.png')

###############################################################################
# .. rst-class:: centered small fst-italic fw-semibold
#
# DKI standard kurtosis measures obtained on unsmoothed data with constrained
# optimization.
#
#
# Mean kurtosis tensor and kurtosis fractional anisotropy
# =======================================================
#
# As pointed by previous studies [Hansen2013]_, axial, radial and mean kurtosis
# depends on the information of both diffusion and kurtosis tensor. DKI
# measures that only depend on the kurtosis tensor include the mean of the
# kurtosis tensor [Hansen2013]_, and the kurtosis fractional anisotropy
# [GlennR2015]_. These measures are computed and illustrated below:

compare_maps([dkifit_plus], ['mkt', 'kfa'], fit_labels=['DKI+'],
             map_kwargs=[{'vmin': 0, 'vmax': 1.5}, {'vmin': 0, 'vmax': 1}],
             filename='Measures_from_kurtosis_tensor_only.png')

###############################################################################
# .. rst-class:: centered small fst-italic fw-semibold
#
# DKI measures obtained from the kurtosis tensor only.
#
#
# As reported by [Hansen2013]_, the mean of the kurtosis tensor (MKT) produces
# similar maps than the standard mean kurtosis (MK). On the other hand,
# the kurtosis fractional anisotropy (KFA) maps shows that the kurtosis tensor
# have different degrees of anisotropy than the FA measures from the diffusion
# tensor.
#
# References
# ----------
# .. [Jensen2005] Jensen JH, Helpern JA, Ramani A, Lu H, Kaczynski K (2005).
#                 Diffusional Kurtosis Imaging: The Quantification of
#                 Non_Gaussian Water Diffusion by Means of Magnetic Resonance
#                 Imaging. Magnetic Resonance in Medicine 53: 1432-1440
# .. [Jensen2010] Jensen JH, Helpern JA (2010). MRI quantification of
#                 non-Gaussian water diffusion by kurtosis analysis. NMR in
#                 Biomedicine 23(7): 698-710
# .. [Fierem2011] Fieremans E, Jensen JH, Helpern JA (2011). White matter
#                 characterization with diffusion kurtosis imaging. NeuroImage
#                 58: 177-188
# .. [Veraar2011] Veraart J, Poot DH, Van Hecke W, Blockx I, Van der Linden A,
#                 Verhoye M, Sijbers J (2011). More Accurate Estimation of
#                 Diffusion Tensor Parameters Using Diffusion Kurtosis Imaging.
#                 Magnetic Resonance in Medicine 65(1): 138-145
# .. [NetoHe2012] Neto Henriques R, Ferreira H, Correia M, (2012). Diffusion
#                 kurtosis imaging of the healthy human brain. Master
#                 Dissertation Bachelor and Master Programin Biomedical
#                 Engineering and Biophysics, Faculty of Sciences.
#                 https://repositorio.ul.pt/bitstream/10451/8511/1/ulfc104137_tm_Rafael_Henriques.pdf
# .. [Hansen2013] Hansen B, Lund TE, Sangill R, and Jespersen SN (2013).
#                 Experimentally and computationally393fast method for
#                 estimation of a mean kurtosis. Magnetic Resonance in
#                 Medicine 69, 1754–1760.394doi:10.1002/mrm.24743
# .. [GlennR2015] Glenn GR, Helpern JA, Tabesh A, Jensen JH (2015).
#                 Quantitative assessment of diffusional387kurtosis anisotropy.
#                 NMR in Biomedicine28, 448–459. doi:10.1002/nbm.3271
# .. [NetoHe2015] Neto Henriques R, Correia MM, Nunes RG, Ferreira HA (2015).
#                 Exploring the 3D geometry of the diffusion kurtosis tensor -
#                 Impact on the development of robust tractography procedures
#                 and novel biomarkers, NeuroImage 111: 85-99
# .. [Perron2015] Perrone D, Aelterman J, Pižurica A, Jeurissen B, Philips W,
#                 Leemans A, (2015). The effect of Gibbs ringing artifacts on
#                 measures derived from diffusion MRI. Neuroimage 120, 441-455.
#                 https://doi.org/10.1016/j.neuroimage.2015.06.068.
# .. [TaxCMW2015] Tax CMW, Otte WM, Viergever MA, Dijkhuizen RM, Leemans A
#                 (2014). REKINDLE: Robust extraction of kurtosis INDices with
#                 linear estimation. Magnetic Resonance in Medicine 73(2):
#                 794-808.
# .. [Hansen2016] Hansen, B, Jespersen, SN (2016). Data for evaluation of fast
#                 kurtosis strategies, b-value optimization and exploration of
#                 diffusion MRI contrast. Scientific Data 3: 160072
#                 doi:10.1038/sdata.2016.72
# .. [NetoHe2018] Neto Henriques R (2018). Advanced Methods for Diffusion MRI
#                 Data Analysis and their Application to the Healthy Ageing
#                 Brain (Doctoral thesis). https://doi.org/10.17863/CAM.29356
# .. [DelaHa2020] Dela Haije et al. "Enforcing necessary non-negativity
#                 constraints for common diffusion MRI models using sum of
#                 squares programming". NeuroImage 209, 2020, 116405.



import pytest
import warnings

import numpy.testing as npt

from dipy.utils.optpkg import optional_package
from dipy.testing.decorators import use_xvfb
from dipy.viz.horizon.tab.base import (build_checkbox, build_label,
                                       build_slider, build_switcher)

fury, has_fury, setup_module = optional_package('fury', min_version="0.9.0")

skip_it = use_xvfb == 'skip'


def check_label(label):
    npt.assert_equal(label.font_family, 'Arial')
    npt.assert_equal(label.justification, 'left')
    npt.assert_equal(label.italic, False)
    npt.assert_equal(label.shadow, False)
    npt.assert_equal(label.actor.GetTextProperty().GetBackgroundColor(),
                     (0., 0., 0.))
    npt.assert_equal(label.actor.GetTextProperty().GetBackgroundOpacity(), 0.0)
    npt.assert_equal(label.color, (0.7, 0.7, 0.7))


@pytest.mark.skipif(skip_it or not has_fury, reason="Needs xvfb")
def test_build_label():
    regular_label = build_label(text='Hello')
    npt.assert_equal(regular_label.message, 'Hello')
    npt.assert_equal(regular_label.font_size, 16)
    npt.assert_equal(regular_label.bold, False)
    check_label(regular_label)

    regular_label = build_label(text='Hello', font_size=10, bold=True)
    npt.assert_equal(regular_label.message, 'Hello')
    npt.assert_equal(regular_label.font_size, 10)
    npt.assert_equal(regular_label.bold, True)
    check_label(regular_label)

    horizon_label = build_label(text='Hello', is_horizon_label=True)
    npt.assert_equal(horizon_label.obj.message, 'Hello')
    npt.assert_equal(horizon_label.obj.font_size, 16)
    npt.assert_equal(horizon_label.obj.bold, False)
    check_label(horizon_label.obj)


@pytest.mark.skipif(skip_it or not has_fury, reason="Needs xvfb")
def test_build_slider():
    single_slider_label, single_slider = build_slider(5, 100)
    npt.assert_equal(single_slider_label.obj.message, '')
    npt.assert_equal(single_slider_label.obj.font_size, 16)
    npt.assert_equal(single_slider_label.obj.bold, False)
    npt.assert_equal(single_slider.obj.value, 5)
    npt.assert_equal(single_slider.obj.max_value, 100)
    npt.assert_equal(single_slider.obj.min_value, 0)
    npt.assert_equal(single_slider.obj.track.width, 450)
    npt.assert_equal(single_slider.obj.track.height, 3)
    npt.assert_equal(single_slider.obj.track.color, (.8, .3, .0))
    npt.assert_equal(single_slider.obj.default_color, (1., .5, .0))
    npt.assert_equal(single_slider.obj.active_color, (.9, .4, .0))
    npt.assert_equal(single_slider.obj.handle.color, (1., .5, .0))
    npt.assert_equal(single_slider.obj.handle.outer_radius, 8)
    npt.assert_equal(single_slider.obj.text.font_size, 16)
    npt.assert_equal(single_slider.obj.text_template,
                     '{value:.1f} ({ratio:.0%})')
    npt.assert_equal(single_slider.selected_value, 5)

    double_slider_label, double_slider = build_slider(
        (4, 5), 100, text_template='{value:.1f}', is_double_slider=True)
    npt.assert_equal(double_slider_label.obj.message, '')
    npt.assert_equal(double_slider_label.obj.font_size, 16)
    npt.assert_equal(double_slider_label.obj.bold, False)
    npt.assert_equal(double_slider.obj.max_value, 100)
    npt.assert_equal(double_slider.obj.min_value, 0)
    npt.assert_equal(double_slider.obj.track.width, 450)
    npt.assert_equal(double_slider.obj.track.height, 3)
    npt.assert_equal(double_slider.obj.track.color, (.8, .3, .0))
    npt.assert_equal(double_slider.obj.default_color, (1., .5, .0))
    npt.assert_equal(double_slider.obj.active_color, (.9, .4, .0))
    npt.assert_equal(double_slider.obj.handles[0].color, (1., .5, .0))
    npt.assert_equal(double_slider.obj.handles[1].color, (1., .5, .0))
    npt.assert_equal(double_slider.obj.handles[0].outer_radius, 8)
    npt.assert_equal(double_slider.obj.handles[1].outer_radius, 8)
    npt.assert_equal(double_slider.obj.text[0].font_size, 16)
    npt.assert_equal(double_slider.obj.text[1].font_size, 16)
    npt.assert_equal(double_slider.selected_value, (4, 5))


def check_for_warnings(warn_printed, w_msg):
    selected_w = [w for w in warn_printed if issubclass(w.category,
                                                        UserWarning)]
    assert len(selected_w) >= 1
    msg = [str(m.message) for m in selected_w]
    npt.assert_equal(w_msg in msg, True)


@pytest.mark.skipif(skip_it or not has_fury, reason="Needs xvfb")
def test_build_checkbox():
    checkbox = build_checkbox(['Hello', 'Hi'], ['Hello'])

    npt.assert_equal(len(checkbox.obj.checked_labels), 1)
    npt.assert_equal(len(checkbox.obj.labels), 2)
    npt.assert_equal(checkbox.selected_value, ['Hello'])

    with warnings.catch_warnings(record=True) as l_warns:
        checkbox = build_checkbox()
        npt.assert_equal(checkbox, None)
        check_for_warnings(l_warns, 'At least one label needs to be to create'
                           + ' checkboxes')

    with warnings.catch_warnings(record=True) as l_warns:
        checkbox = build_checkbox([])
        npt.assert_equal(checkbox, None)
        check_for_warnings(l_warns, 'At least one label needs to be to create'
                           + ' checkboxes')


@pytest.mark.skipif(skip_it or not has_fury, reason="Needs xvfb")
def test_build_switcher():
    with warnings.catch_warnings(record=True) as l_warns:
        none_switcher = build_switcher()
        npt.assert_equal(none_switcher, None)
        check_for_warnings(l_warns, 'No items passed in switcher')

    switcher_label, switcher = build_switcher(
        [{'label': 'Hello', 'value': 'hello'}])
    npt.assert_equal(switcher.selected_value[0], 0)
    npt.assert_equal(switcher.selected_value[1], 'hello')
    npt.assert_equal(switcher_label.selected_value, '')

    switcher_label, switcher = build_switcher(
        [{'label': 'Hello', 'value': 'hello'}], 'Greeting', 1)
    npt.assert_equal(switcher.selected_value[0], 0)
    npt.assert_equal(switcher.selected_value[1], 'hello')
    npt.assert_equal(switcher_label.selected_value, 'Greeting')



import numpy as np

from dipy.utils.optpkg import optional_package
from dipy.viz.horizon.tab import (HorizonTab, build_label, color_single_slider)

fury, has_fury, setup_module = optional_package('fury', min_version="0.9.0")

if has_fury:
    from fury import ui


class ClustersTab(HorizonTab):
    def __init__(self, clusters_visualizer, threshold):
        self.__visualizer = clusters_visualizer

        self.__centroid_actors = self.__visualizer.centroid_actors
        self.__cluster_actors = self.__visualizer.cluster_actors
        self.__name = 'Clusters'

        self.__tab_id = 0
        self.__tab_ui = None

        length = 450
        lw = 3
        radius = 8
        fs = 16
        tt = '{value:.0f}'

        self.__slider_label_size = build_label(text='Size')

        sizes = self.__visualizer.sizes
        self.__selected_size = np.percentile(sizes, 50)

        self.__slider_size = ui.LineSlider2D(
            initial_value=self.__selected_size, min_value=np.min(sizes),
            max_value=np.percentile(sizes, 98), length=length, line_width=lw,
            outer_radius=radius, font_size=fs, text_template=tt)

        color_single_slider(self.__slider_size)

        self.__slider_size.handle_events(self.__slider_size.handle.actor)
        self.__slider_size.on_left_mouse_button_released = self.__change_size

        self.__slider_label_length = build_label(text='Length')

        lengths = self.__visualizer.lengths
        self.__selected_length = np.percentile(lengths, 25)

        self.__slider_length = ui.LineSlider2D(
            initial_value=self.__selected_length, min_value=np.min(lengths),
            max_value=np.percentile(lengths, 98), length=length, line_width=lw,
            outer_radius=radius, font_size=fs, text_template=tt)

        color_single_slider(self.__slider_length)

        self.__slider_length.handle_events(self.__slider_length.handle.actor)
        self.__slider_length.on_left_mouse_button_released = (
            self.__change_length)

        self.__slider_label_threshold = build_label(text='Threshold')

        self.__selected_threshold = threshold

        self.__slider_threshold = ui.LineSlider2D(
            initial_value=self.__selected_threshold, min_value=5, max_value=25,
            length=length, line_width=lw, outer_radius=radius, font_size=fs,
            text_template=tt)

        color_single_slider(self.__slider_threshold)

        self.__slider_threshold.handle_events(
            self.__slider_threshold.handle.actor)
        self.__slider_threshold.on_left_mouse_button_released = (
            self.__change_threshold)

    def __change_length(self, istyle, obj, slider):
        self.__selected_length = int(np.rint(slider.value))
        self.__update_clusters()
        istyle.force_render()

    def __change_size(self, istyle, obj, slider):
        self.__selected_size = int(np.rint(slider.value))
        self.__update_clusters()
        istyle.force_render()

    def __change_threshold(self, istyle, obj, slider):
        value = int(np.rint(slider.value))
        if value != self.__selected_threshold:
            self.__visualizer.recluster_tractograms(value)

            sizes = self.__visualizer.sizes
            self.__selected_size = np.percentile(sizes, 50)

            lengths = self.__visualizer.lengths
            self.__selected_length = np.percentile(lengths, 25)

            self.__update_clusters()

            # Updating size slider
            self.__slider_size.min_value = np.min(sizes)
            self.__slider_size.max_value = np.percentile(sizes, 98)
            self.__slider_size.value = self.__selected_size
            self.__slider_size.update()

            # Updating length slider
            self.__slider_length.min_value = np.min(lengths)
            self.__slider_length.max_value = np.percentile(lengths, 98)
            self.__slider_length.value = self.__selected_length
            self.__slider_length.update()

            self.__selected_threshold = value
            istyle.force_render()

    def __update_clusters(self):
        for k in self.__cluster_actors:
            length_validation = (
                self.__cluster_actors[k]['length'] < self.__selected_length)
            size_validation = (
                self.__cluster_actors[k]['size'] < self.__selected_size)
            if (length_validation or size_validation):
                self.__cluster_actors[k]['actor'].SetVisibility(False)
                if k.GetVisibility():
                    k.SetVisibility(False)
            else:
                self.__cluster_actors[k]['actor'].SetVisibility(True)

    def build(self, tab_id, tab_ui):
        self.__tab_id = tab_id
        self.__tab_ui = tab_ui

        x_pos = .02

        self.__tab_ui.add_element(
            self.__tab_id, self.__slider_label_size, (x_pos, .85))
        self.__tab_ui.add_element(
            self.__tab_id, self.__slider_label_length, (x_pos, .62))
        self.__tab_ui.add_element(
            self.__tab_id, self.__slider_label_threshold, (x_pos, .38))

        x_pos=.12

        self.__tab_ui.add_element(
            self.__tab_id, self.__slider_size, (x_pos, .85))
        self.__tab_ui.add_element(
            self.__tab_id, self.__slider_length, (x_pos, .62))
        self.__tab_ui.add_element(
            self.__tab_id, self.__slider_threshold, (x_pos, .38))

    @property
    def name(self):
        return self.__name



import pytest

from dipy.utils.optpkg import optional_package
from dipy.utils.tripwire import TripWireError
from dipy.testing import assert_true, assert_false


def test_optional_package():
    pkg, have_pkg, setup_module = optional_package('os')
    assert_true(have_pkg)
    assert_true(hasattr(pkg, 'path'))

    pkg, have_pkg, setup_module = optional_package('not_a_package')
    assert_false(have_pkg)
    with pytest.raises(TripWireError):
        pkg.some_function()

    pkg, have_pkg, setup_module = optional_package('dipy', min_version='10.0.0')
    assert_false(have_pkg)
    with pytest.raises(TripWireError):
        pkg.some_function()

    pkg, have_pkg, setup_module = optional_package('dipy', min_version='1.0.0')
    assert_true(have_pkg)



import numpy as np
from scipy.ndimage import affine_transform, label
from dipy.align.reslice import reslice


def normalize(image, min_v=None, max_v=None, new_min=-1, new_max=1):
    r"""
    normalization function

    Parameters
    ----------
    image : np.ndarray
    min_v : int or float, optional
        minimum value range for normalization
        intensities below min_v will be clipped
        if None it is set to min value of image
        Default : None
    max_v : int or float, optional
        maximum value range for normalization
        intensities above max_v will be clipped
        if None it is set to max value of image
        Default : None
    new_min : int or float, optional
        new minimum value after normalization
        Default : 0
    new_max : int or float, optional
        new maximum value after normalization
        Default : 1

    Returns
    -------
    np.ndarray
        Normalized image from range new_min to new_max
    """
    if min_v is None:
        min_v = np.min(image)
    if max_v is None:
        max_v = np.max(image)
    return np.interp(image, (min_v, max_v), (new_min, new_max))


def unnormalize(image, norm_min, norm_max, min_v, max_v):
    r"""
    unnormalization function

    Parameters
    ----------
    image : np.ndarray
    norm_min : int or float
        minimum value of normalized image
    norm_max : int or float
        maximum value of normalized image
    min_v : int or float
        minimum value of unnormalized image
    max_v : int or float
        maximum value of unnormalized image

    Returns
    -------
    np.ndarray
        unnormalized image from range min_v to max_v
    """
    return (image - norm_min) / (norm_max-norm_min) * \
           (max_v - min_v) + min_v


def set_logger_level(log_level, logger):
    """ Change the logger to one of the following:
    DEBUG, INFO, WARNING, CRITICAL, ERROR

    Parameters
    ----------
    log_level : str
        Log level for the logger
    """
    logger.setLevel(level=log_level)


def transform_img(image, affine, voxsize=None,
                  init_shape=(256, 256, 256), scale=2):
    r"""
    Function to reshape image as an input to the model

    Parameters
    ----------
    image : np.ndarray
        Image to transform to voxelspace
    affine : np.ndarray
        Affine matrix provided by the file
    voxsize : np.ndarray (3,), optional
        Voxel size of the image
    init_shape : tuple, optional
        Initial shape to transform the image to
        Default is (256, 256, 256)
    scale : float, optional
        How much we want to scale the image
        Default is 2

    Returns
    -------
    transformed_img : np.ndarray
    """
    if voxsize is not None and np.any(voxsize != np.ones(3)):
        image, affine2 = reslice(image, affine, voxsize, (1, 1, 1))
    else:
        affine2 = affine.copy()
    ori_shape = np.array(image.shape)
    affine2[:3, 3] += np.array([init_shape[0]//2,
                                init_shape[1]//2,
                                init_shape[2]//2])
    inv_affine = np.linalg.inv(affine2)
    transformed_img = affine_transform(image, inv_affine, output_shape=init_shape)
    transformed_img, _ = reslice(transformed_img, np.eye(4), (1, 1, 1),
                                 (scale, scale, scale))
    return transformed_img, affine2, ori_shape


def recover_img(image, affine, ori_shape, image_shape,
                init_shape=(256, 256, 256), voxsize=None, scale=2):
    r"""
    Function to recover image back to its original shape

    Parameters
    ----------
    image : np.ndarray
        Image to recover
    affine : np.ndarray
        Affine matrix provided from transform_img
    ori_shape : np.ndarray (3,)
        Original shape of isotropic image
    image_shape : tuple (3,)
        Original shape of actual image
    init_shape : tuple (3,), optional
        Initial shape to transform the image to
        Default is (256, 256, 256)
    voxsize : np.ndarray (3,), optional
        Voxel size of the original image
    scale : float, optional
        Scale that was used in transform_img
        Default is 2

    Returns
    -------
    recovered_img : np.ndarray
    """
    new_image, _ = reslice(image, np.eye(4), (scale, scale, scale), (1, 1, 1))
    recovered_img = affine_transform(new_image, affine, output_shape=ori_shape)
    affine[:3, 3] += np.array([init_shape[0]//2,
                               init_shape[1]//2,
                               init_shape[2]//2])
    if voxsize is not None and np.any(voxsize != np.ones(3)):
        kwargs = {'order': 1,
                  'mode': 'constant',
                  'matrix': voxsize,
                  'output_shape': image_shape}
        recovered_img = np.round(affine_transform(recovered_img, **kwargs))
    return recovered_img


def correct_minor_errors(binary_img):
    """
    Remove any small mask chunks or holes
    that could be in the segmentation output.

    Parameters
    ----------
    binary_img : np.ndarray
        Binary image

    Returns
    -------
    largest_img : np.ndarray
    """
    largest_img = np.zeros_like(binary_img)
    chunks, n_chunk = label(np.abs(1-binary_img))
    u, c = np.unique(chunks[chunks != 0], return_counts=True)
    target = u[np.argmax(c)]
    largest_img = np.where(chunks == target, 0, 1)

    chunks, n_chunk = label(largest_img)
    u, c = np.unique(chunks[chunks != 0], return_counts=True)
    target = u[np.argmax(c)]
    largest_img = np.where(chunks == target, 1, 0)

    for x in range(largest_img.shape[0]):
        chunks, n_chunk = label(np.abs(1-largest_img[x]))
        u, c = np.unique(chunks[chunks != 0], return_counts=True)
        target = u[np.argmax(c)]
        largest_img[x] = np.where(chunks == target, 0, 1)
    for y in range(largest_img.shape[1]):
        chunks, n_chunk = label(np.abs(1-largest_img[:, y]))
        u, c = np.unique(chunks[chunks != 0], return_counts=True)
        target = u[np.argmax(c)]
        largest_img[:, y] = np.where(chunks == target, 0, 1)
    for z in range(largest_img.shape[2]):
        chunks, n_chunk = label(np.abs(1-largest_img[..., z]))
        u, c = np.unique(chunks[chunks != 0], return_counts=True)
        target = u[np.argmax(c)]
        largest_img[..., z] = np.where(chunks == target, 0, 1)

    return largest_img



"""
=================================================================
Continuous and analytical diffusion signal modelling with MAP-MRI
=================================================================

We show how to model the diffusion signal as a linear combination of continuous
functions from the MAP-MRI basis [Ozarslan2013]_. This continuous
representation allows for the computation of many properties of both the signal
and diffusion propagator.

We show how to estimate the analytical orientation distribution function (ODF)
and a variety of scalar indices. These include rotationally invariant
quantities such as the mean squared displacement (MSD), Q-space inverse
variance (QIV), teturn-to-origin probability (RTOP) and non-Gaussianity (NG).
Interestingly, the MAP-MRI basis also allows for the computation of directional
indices, such as the return-to-axis probability (RTAP), the return-to-plane
probability (RTPP), and the parallel and perpendicular non-Gaussianity.

The estimation of these properties from noisy and sparse DWIs requires that the
fitting of the MAP-MRI basis is constrained and/or regularized. This can be
done by constraining the diffusion propagator to positive values
[Ozarslan2013]_ [DelaHaije2020]_, and through analytic Laplacian
regularization (MAPL) [Fick2016a]_.

First import the necessary modules:
"""

from dipy.reconst import mapmri
from dipy.viz import window, actor
from dipy.data import get_fnames, get_sphere
from dipy.core.gradients import gradient_table
from dipy.io.image import load_nifti
from dipy.io.gradients import read_bvals_bvecs
from dipy.viz.plotting import compare_maps

###############################################################################
# Download and read the data for this tutorial.
#
# MAP-MRI requires multi-shell data, to properly fit the radial part of the
# basis. The total size of the downloaded data is 187.66 MBytes, however you
# only need to fetch it once.

fraw, fbval, fbvec, t1_fname = get_fnames('cfin_multib')

###############################################################################
# ``data`` contains the voxel data and ``gtab`` contains a ``GradientTable``
# object (gradient information e.g. b-values). For example, to show the
# b-values it is possible to write::
#
#    print(gtab.bvals)
#
# For the values of the q-space indices to make sense it is necessary to
# explicitly state the ``big_delta`` and ``small_delta`` parameters in the
# gradient table.

data, affine = load_nifti(fraw)
bvals, bvecs = read_bvals_bvecs(fbval, fbvec)
gtab = gradient_table(bvals, bvecs)

big_delta = 0.0365  # seconds
small_delta = 0.0157  # seconds
gtab = gradient_table(bvals=gtab.bvals, bvecs=gtab.bvecs,
                      big_delta=big_delta,
                      small_delta=small_delta)

data_small = data[40:65, 50:51]

print('data.shape (%d, %d, %d, %d)' % data.shape)

###############################################################################
# The MAP-MRI Model can now be instantiated. The ``radial_order`` determines
# the expansion order of the basis, i.e., how many basis functions are used to
# approximate the signal.
#
# First, we must decide to use the anisotropic or isotropic MAP-MRI basis. As
# was shown in [Fick2016a]_, the isotropic basis is best used for tractography
# purposes, as the anisotropic basis has a bias towards smaller crossing angles
# in the ODF. For signal fitting and estimation of scalar quantities the
# anisotropic basis is suggested. The choice can be made by setting
# ``anisotropic_scaling=True`` or ``anisotropic_scaling=False``.
#
# First, we must select the method of regularization and/or constraining the
# basis fitting.
#
# - ``laplacian_regularization=True`` makes it use Laplacian regularization
#   [Fick2016a]_. This method essentially reduces spurious oscillations in the
#   reconstruction by minimizing the Laplacian of the fitted signal. Several
#   options can be given to select the regularization weight:
#
#     - ``regularization_weighting=GCV`` uses generalized cross-validation
#       [Craven1978]_ to find an optimal weight.
#     - ``regularization_weighting=0.2`` for example would omit the GCV and
#       just set it to 0.2 (found to be reasonable in HCP data [Fick2016a]_).
#     - ``regularization_weighting=np.array(weights)`` would make the GCV use
#       a custom range to find an optimal weight.
#
# - ``positivity_constraint=True`` makes it use a positivity constraint on the
#   diffusion propagator. This method constrains the final solution of the
#   diffusion propagator to be positive either globally [DelaHaije2020] or at a
#   set of discrete points [Ozarslan2013]_, since negative values should not
#   exist.
#
#     - Setting ``global_constraints=True`` enforces positivity everywhere.
#       With the setting ``global_constraints=False`` positivity is enforced on
#       a grid determined by ``pos_grid`` and ``pos_radius``.
#
# Both methods do a good job of producing viable solutions to the signal
# fitting in practice. The difference is that the Laplacian regularization
# imposes smoothness over the entire signal, including the extrapolation
# beyond the measured signal. In practice this may result in, but does not
# guarantee, positive solutions of the diffusion propagator. The positivity
# constraint guarantees a positive solution which in general results in smooth
# solutions, but does not guarantee it.
#
# A suggested strategy is to use a low Laplacian weight together with a
# positivity constraint. In this way both desired properties are guaranteed
# in final solution. Higher Laplacian weights are recommended when the data is
# very noisy.
#
# We use the package CVXPY (https://www.cvxpy.org/) to solve convex
# optimization problems when ``positivity_constraint=True``, so we need to
# first install CVXPY. When using ``global_constraints=True`` to ensure
# global positivity, it is recommended to use the MOSEK solver
# (https://www.mosek.com/) together with CVXPY by setting
# ``cvxpy_solver='MOSEK'``. Different solvers can differ greatly in
# terms of runtime and solution accuracy, and in some cases solvers may show
# warnings about convergence or recommended option settings.
#
# For now we will generate the anisotropic models for different combinations.

radial_order = 6

map_model_laplacian_aniso = mapmri.MapmriModel(gtab, radial_order=radial_order,
                                               laplacian_regularization=True,
                                               laplacian_weighting=.2)

map_model_positivity_aniso = mapmri.MapmriModel(gtab,
                                                radial_order=radial_order,
                                                laplacian_regularization=False,
                                                positivity_constraint=True)

map_model_both_aniso = mapmri.MapmriModel(gtab, radial_order=radial_order,
                                          laplacian_regularization=True,
                                          laplacian_weighting=.05,
                                          positivity_constraint=True)

map_model_plus_aniso = mapmri.MapmriModel(gtab,
                                          radial_order=radial_order,
                                          laplacian_regularization=False,
                                          positivity_constraint=True,
                                          global_constraints=True)

###############################################################################
# Note that when we use only Laplacian regularization, the ``GCV`` option may
# select very low regularization weights in very anisotropic white matter such
# as the corpus callosum, resulting in corrupted scalar indices. In this
# example we therefore choose a preset value of 0.2, which was shown to be
# quite robust and also faster in practice [Fick2016a]_.
#
# We can then fit the MAP-MRI model to the data:

mapfit_laplacian_aniso = map_model_laplacian_aniso.fit(data_small)
mapfit_positivity_aniso = map_model_positivity_aniso.fit(data_small)
mapfit_both_aniso = map_model_both_aniso.fit(data_small)
mapfit_plus_aniso = map_model_plus_aniso.fit(data_small)

fits = [mapfit_laplacian_aniso, mapfit_positivity_aniso, mapfit_both_aniso,
        mapfit_plus_aniso]
fit_labels = ['MAPL', 'CMAP', 'CMAPL', 'MAP+']

###############################################################################
# From the fitted models we will first illustrate the estimation of q-space
# indices. We will compare the estimation using only Laplacian regularization
# (MAPL), using only the global positivity constraint (MAP+), using only
# positivity in discrete points (CMAP), or using both Laplacian regularization
# and positivity in discrete points (CMAPL). We first show the RTOP
# [Ozarslan2013]_.

compare_maps(fits, maps=['rtop'], fit_labels=fit_labels, map_labels=['RTOP'],
             filename='MAPMRI_rtop.png')

###############################################################################
# .. rst-class:: centered small fst-italic fw-semibold
#
# It can be seen that all maps appear quite smooth and similar, although it is
# clear that the global positivity constraints provide smoother maps than the
# discrete constraints. Higher Laplacian weights also produce smoother maps,
# but tend to suppress the estimated RTOP values. The similarity and
# differences in reconstruction can be further illustrated by visualizing the
# analytic norm of the Laplacian of the fitted signal.

compare_maps(fits, maps=['norm_of_laplacian_signal'], fit_labels=fit_labels,
             map_labels=['Norm of Laplacian'],
             map_kwargs={'vmin': 0, 'vmax': 3},
             filename='MAPMRI_norm_laplacian.png')

###############################################################################
# .. rst-class:: centered small fst-italic fw-semibold
#
# A high Laplacian norm indicates that the gradient in the three-dimensional
# signal reconstruction changes a lot -- something that may indicate spurious
# oscillations. In the Laplacian reconstruction (left) we see that there are
# some isolated voxels that have a higher norm than the rest. In the positivity
# constrained reconstruction the norm is already smoother. When both methods
# are used together the overall norm gets smoother still, since both smoothness
# of the signal and positivity of the propagator are imposed.
#
# From now on we will just use the combined approach and the globally
# constrained approach, show all maps we can generate, and explain their
# significance.

fits = fits[2:]
fit_labels = fit_labels[2:]

compare_maps(fits, maps=['msd', 'qiv', 'rtop', 'rtap', 'rtpp'],
             fit_labels=fit_labels,
             map_labels=['MSD', 'QIV', 'RTOP', 'RTAP', 'RTPP'],
             filename='MAPMRI_maps.png')

###############################################################################
# .. rst-class:: centered small fst-italic fw-semibold
#
# From left to right:
#
# - Mean squared displacement (MSD) is a measure for how far protons are able
#   to diffuse. a decrease in MSD indicates protons are hindered/restricted
#   more, as can be seen by the high MSD in the CSF, but low in the white
#   matter.
# - Q-space Inverse Variance (QIV) is a measure of variance in the signal,
#   which is said to have higher contrast to white matter than the MSD
#   [Hosseinbor2013]_. QIV has also been shown to have high sensitivity to
#   tissue composition change in a simulation study [Fick2016b]_.
# - Return-to-origin probability (RTOP) quantifies the probability that a
#   proton will be at the same position at the first and second diffusion
#   gradient pulse. A higher RTOP indicates that the volume a spin is inside
#   of is smaller, meaning more overall restriction. This is illustrated by
#   the low values in CSF and high values in white matter.
# - Return-to-axis probability (RTAP) is a directional index that quantifies
#   the probability that a proton will be along the axis of the main
#   eigenvector of a diffusion tensor during both diffusion gradient pulses.
#   RTAP has been related to the apparent axon diameter [Ozarslan2013]_
#   [Fick2016a]_ under several strong assumptions on the tissue composition
#   and acquisition protocol.
# - Return-to-plane probability (RTPP) is a directional index that quantifies
#   the probability that a proton will be on the plane perpendicular to the
#   main eigenvector of a diffusion tensor during both gradient pulses. RTPP
#   is related to the length of a pore [Ozarslan2013]_ but in practice should
#   be similar to that of Gaussian diffusion.
#
#
# It is also possible to estimate the amount of non-Gaussian diffusion in every
# voxel [Ozarslan2013]_. This quantity is estimated through the ratio between
# Gaussian volume (MAP-MRI's first basis function) and the non-Gaussian volume
# (all other basis functions) of the fitted signal. For this value to be
# physically meaningful we must use a b-value threshold in the MAP-MRI model.
# This threshold makes the scale estimation in MAP-MRI use only samples that
# realistically describe Gaussian diffusion, i.e., at low b-values.

map_model_both_ng = mapmri.MapmriModel(gtab, radial_order=radial_order,
                                       laplacian_regularization=True,
                                       laplacian_weighting=.05,
                                       positivity_constraint=True,
                                       bval_threshold=2000)

map_model_plus_ng = mapmri.MapmriModel(gtab, radial_order=radial_order,
                                       positivity_constraint=True,
                                       global_constraints=True,
                                       bval_threshold=2000)

mapfit_both_ng = map_model_both_ng.fit(data_small)
mapfit_plus_ng = map_model_plus_ng.fit(data_small)

fits = [mapfit_both_ng, mapfit_plus_ng]
fit_labels = ['CMAPL', 'MAP+']

compare_maps(fits, maps=['ng', 'ng_perpendicular', 'ng_parallel'],
             fit_labels=fit_labels,
             map_labels=['NG', 'NG perpendicular', 'NG parallel'],
             filename='MAPMRI_ng.png')

###############################################################################
# .. rst-class:: centered small fst-italic fw-semibold
#
# On the left we see the overall NG and on the right the directional
# perpendicular NG and parallel NG. The NG ranges from 1
# (completely non-Gaussian) to 0 (completely Gaussian). The overall NG of a
# voxel is always higher or equal than each of its components. It can be seen
# that NG has low values in the CSF and higher in the white matter.
#
# Increases or decreases in these values do not point to a specific
# microstructural change, but can indicate clues as to what is happening,
# similar to fractional anisotropy. An initial simulation study that quantifies
# the added value of q-space indices over DTI indices is given in [Fick2016b]_.
#
# The MAP-MRI framework also allows for the estimation of orientation
# distribution functions (ODFs). We recommend to use the isotropic
# implementation for this purpose, as the anisotropic implementation has a bias
# towards smaller crossing angles.
#
# For the isotropic basis we recommend to use a ``radial_order`` of 8, as the
# basis needs more generic and needs more basis functions to approximate the
# signal.

radial_order = 8
map_model_both_iso = mapmri.MapmriModel(gtab, radial_order=radial_order,
                                        laplacian_regularization=True,
                                        laplacian_weighting=.1,
                                        positivity_constraint=True,
                                        anisotropic_scaling=False)

mapfit_both_iso = map_model_both_iso.fit(data_small)

###############################################################################
# Load an ODF reconstruction sphere.

sphere = get_sphere('repulsion724')

###############################################################################
# Compute the ODFs.
#
# The radial order ``s`` can be increased to sharpen the results, but it might
# also make the ODFs noisier. Always check the results visually.

odf = mapfit_both_iso.odf(sphere, s=2)
print('odf.shape (%d, %d, %d, %d)' % odf.shape)

###############################################################################
# Display the ODFs.

# Enables/disables interactive visualization
interactive = False

scene = window.Scene()
sfu = actor.odf_slicer(odf, sphere=sphere, colormap='plasma', scale=0.5)
sfu.display(y=0)
sfu.RotateX(-90)
scene.add(sfu)
window.record(scene, out_path='odfs.png', size=(600, 600))
if interactive:
    window.show(scene)

###############################################################################
# .. rst-class:: centered small fst-italic fw-semibold
#
# Orientation distribution functions (ODFs).
#
#
# References
# ----------
#
# .. [Ozarslan2013] Ozarslan E. et al., "Mean apparent propagator (MAP) MRI: A
#    novel diffusion imaging method for mapping tissue microstructure",
#    NeuroImage, 2013.
#
# .. [Fick2016a] Fick, Rutger HJ, et al. "MAPL: Tissue microstructure
#    estimation using Laplacian-regularized MAP-MRI and its application to HCP
#    data." NeuroImage (2016).
#
# .. [Craven1978] Craven et al. "Smoothing Noisy Data with Spline Functions."
#    NUMER MATH 31.4 (1978): 377-403.
#
# .. [Hosseinbor2013] Hosseinbor et al. "Bessel fourier orientation
#    reconstruction (bfor): an analytical diffusion propagator reconstruction
#    for hybrid diffusion imaging and computation of q-space indices.
#    NeuroImage 64, 650-670.
#
# .. [Fick2016b] Fick et al. "A sensitivity analysis of Q-space indices with
#    respect to changes in axonal diameter, dispersion and tissue composition.
#    ISBI 2016.
#
# .. [DelaHaije2020] Dela Haije et al. "Enforcing necessary non-negativity
#    constraints for common diffusion MRI models using sum of squares
#    programming". NeuroImage 209, 2020, 116405.



# cython: boundscheck=False
# cython: initializedcheck=False
# cython: wraparound=False

cdef int where_to_insert(cnp.float_t* arr, cnp.float_t number, int size) nogil:
    cdef:
        int idx
        cnp.float_t current
    for idx in range(size - 1, -1, -1):
        current = arr[idx]
        if number >= current:
            return idx + 1

    return 0


cdef void cumsum(cnp.float_t* arr_in, cnp.float_t* arr_out, int N) nogil:
    cdef:
        int i = 0
        cnp.float_t csum = 0
    for i in range(N):
        csum += arr_in[i]
        arr_out[i] = csum


cdef void copy_point(double * a, double * b) nogil:
    cdef:
        int i = 0
    for i in range(3):
        b[i] = a[i]


cdef void scalar_muliplication_point(double * a, double scalar) nogil:
    cdef:
        int i = 0
    for i in range(3):
        a[i] *= scalar


cdef double norm(double * v) nogil:
    """Compute the vector norm.

    Parameters
    ----------
    v : double[3]
        input vector.

    """
    return sqrt(v[0] * v[0] + v[1] * v[1] + v[2] * v[2])


cdef double dot(double * v1, double * v2) nogil:
    """Compute vectors dot product.

    Parameters
    ----------
    v1 : double[3]
        input vector 1.
    v2 : double[3]
        input vector 2.

    Returns
    -------
    _ : double
        dot product of input vectors.
    """
    return v1[0] * v2[0] + v1[1] * v2[1] + v1[2] * v2[2]


cdef void normalize(double * v) nogil:
    """Normalize the vector.

    Parameters
    ----------
    v : double[3]
        input vector

    Notes
    -----
    Overwrites the first argument.

    """
    cdef double scale = 1.0 / norm(v)
    v[0] = v[0] * scale
    v[1] = v[1] * scale
    v[2] = v[2] * scale


cdef void cross(double * out, double * v1, double * v2) nogil:
    """Compute vectors cross product.

    Parameters
    ----------
    out : double[3]
        output vector.
    v1 : double[3]
        input vector 1.
    v2 : double[3]
        input vector 2.

    Notes
    -----
    Overwrites the first argument.

    """
    out[0] = v1[1] * v2[2] - v1[2] * v2[1]
    out[1] = v1[2] * v2[0] - v1[0] * v2[2]
    out[2] = v1[0] * v2[1] - v1[1] * v2[0]


cdef void random_vector(double * out) nogil:
    """Generate a unit random vector

    Parameters
    ----------
    out : double[3]
        output vector

    Notes
    -----
    Overwrites the input
    """
    out[0] = 2.0 * random() - 1.0
    out[1] = 2.0 * random() - 1.0
    out[2] = 2.0 * random() - 1.0
    normalize(out)


cdef void random_perpendicular_vector(double * out, double * v) nogil:
    """Generate a random perpendicular vector

    Parameters
    ----------
    out : double[3]
        output vector

    v : double[3]
        input vector

    Notes
    -----
    Overwrites the first argument
    """
    cdef double[3] tmp

    random_vector(tmp)
    cross(out, v, tmp)
    normalize(out)


cpdef (double, double) random_point_within_circle(double r) nogil:
    """Generate a random point within a circle

    Parameters
    ----------
    r : double
        The radius of the circle

    Returns
    -------
    x : double
        x coordinate of the random point

    y : double
        y coordinate of the random point

    """
    cdef double x = 1.0
    cdef double y = 1.0

    while (x * x + y * y) > 1:
        x = 2.0 * random() - 1.0
        y = 2.0 * random() - 1.0
    return (r * x, r * y)


cpdef double random() nogil:
    """Sample a random number between 0 and 1.

    Returns
    -------
    _ : double
        random number.
    """
    return rand() / float(RAND_MAX)


cpdef void seed(cnp.npy_uint32 s) nogil:
    """Set the random seed of stdlib.

    Parameters
    ----------
    s : int
        random seed.
    """
    srand(s)



from os.path import join as pjoin
from tempfile import TemporaryDirectory

import numpy as np
import numpy.testing as npt
import pytest

from dipy.data import get_fnames
from dipy.io.image import load_nifti_data, save_nifti
from dipy.nn.evac import EVACPlus
from dipy.utils.optpkg import optional_package
from dipy.workflows.nn import EVACPlusFlow


tf, have_tf, _ = optional_package('tensorflow', min_version='2.0.0')


@pytest.mark.skipif(not have_tf, reason='Requires TensorFlow')
def test_evac_plus_flow():
    with TemporaryDirectory() as out_dir:
        file_path = get_fnames('evac_test_data')

        volume = np.load(file_path)['input'][0]
        temp_affine = np.eye(4)
        temp_path = pjoin(out_dir, 'temp.nii.gz')
        save_nifti(temp_path, volume, temp_affine)
        save_masked = True

        evac_flow = EVACPlusFlow()
        evac_flow.run(temp_path, out_dir=out_dir, save_masked=save_masked)

        mask_name = evac_flow.last_generated_outputs['out_mask']
        masked_name = evac_flow.last_generated_outputs['out_masked']

        evac = EVACPlus()
        mask = evac.predict(volume, temp_affine)
        masked = volume * mask

        result_mask_data = load_nifti_data(pjoin(out_dir, mask_name))
        npt.assert_array_equal(result_mask_data.astype(np.uint8), mask)

        result_masked_data = load_nifti_data(pjoin(out_dir, masked_name))

        npt.assert_array_equal(result_masked_data, masked)



from functools import reduce

import numpy as np
import numpy.testing as npt

from dipy.reconst.multi_voxel import _squash, multi_voxel_fit, CallableArray
from dipy.core.sphere import unit_icosahedron
from dipy.testing.decorators import set_random_number_generator


def test_squash():
    A = np.ones((3, 3), dtype=float)
    B = np.asarray(A, object)
    npt.assert_array_equal(A, _squash(B))
    npt.assert_equal(_squash(B).dtype, A.dtype)

    B[2, 2] = None
    A[2, 2] = 0
    npt.assert_array_equal(A, _squash(B))
    npt.assert_equal(_squash(B).dtype, A.dtype)

    for ijk in np.ndindex(*B.shape):
        B[ijk] = np.ones((2,))
    A = np.ones((3, 3, 2))
    npt.assert_array_equal(A, _squash(B))
    npt.assert_equal(_squash(B).dtype, A.dtype)

    B[2, 2] = None
    A[2, 2] = 0
    npt.assert_array_equal(A, _squash(B))
    npt.assert_equal(_squash(B).dtype, A.dtype)

    # sub-arrays have different shapes ( (3,) and (2,) )
    B[0, 0] = np.ones((3,))
    npt.assert_(_squash(B) is B)

    # Check dtypes for arrays and scalars
    arr_arr = np.zeros((2,), dtype=object)
    scalar_arr = np.zeros((2,), dtype=object)
    numeric_types = [getattr(np, dtype) for dtype in (
        'int8', 'byte', 'int16', 'short', 'int32', 'intc', 'int_', 'int64',
        'longlong', 'uint8', 'ubyte', 'uint16', 'ushort', 'uint32', 'uintc',
        'uint', 'uint64', 'ulonglong', 'float16', 'half', 'float32', 'single',
        'float64', 'double', 'float96', 'float128', 'longdouble', 'complex64',
        'csingle', 'complex128', 'cdouble', 'complex192', 'complex256',
        'clongdouble') if hasattr(np, dtype)] + [bool]
    for dt0 in numeric_types:
        arr_arr[0] = np.zeros((3,), dtype=dt0)
        scalar_arr[0] = dt0(0)
        for dt1 in numeric_types:
            arr_arr[1] = np.zeros((3,), dtype=dt1)
            npt.assert_equal(_squash(arr_arr).dtype,
                             reduce(np.add, arr_arr).dtype)
            scalar_arr[1] = dt0(1)
            npt.assert_equal(_squash(scalar_arr).dtype,
                             reduce(np.add, scalar_arr).dtype)

    # Check masks and Nones
    arr = np.ones((3, 4), dtype=float)
    obj_arr = arr.astype(object)
    arr[1, 1] = 99
    obj_arr[1, 1] = None
    npt.assert_array_equal(_squash(obj_arr, mask=None, fill=99), arr)
    msk = arr == 1
    npt.assert_array_equal(_squash(obj_arr, mask=msk, fill=99), arr)
    msk[1, 1] = 1  # unmask None - object array back
    npt.assert_array_equal(_squash(obj_arr, mask=msk, fill=99), obj_arr)
    msk[1, 1] = 0  # remask, back to fill again
    npt.assert_array_equal(_squash(obj_arr, mask=msk, fill=99), arr)
    obj_arr[2, 3] = None  # add another unmasked None, object again
    npt.assert_array_equal(_squash(obj_arr, mask=msk, fill=99), obj_arr)

    # Check array of arrays
    obj_arrs = np.zeros((3,), dtype=object)
    for i in range(3):
        obj_arrs[i] = np.ones((4, 5))
    arr_arrs = np.ones((3, 4, 5))
    # No Nones
    npt.assert_array_equal(_squash(obj_arrs, mask=None, fill=99), arr_arrs)
    # None, implicit masking
    obj_masked = obj_arrs.copy()
    obj_masked[1] = None
    arr_masked = arr_arrs.copy()
    arr_masked[1] = 99
    npt.assert_array_equal(_squash(obj_masked, mask=None, fill=99),
                           arr_masked)
    msk = np.array([1, 0, 1], dtype=bool)  # explicit mask
    npt.assert_array_equal(_squash(obj_masked, mask=msk, fill=99),
                           arr_masked)


def test_CallableArray():
    callarray = CallableArray((2, 3), dtype=object)

    # Test without Nones
    callarray[:] = np.arange
    expected = np.empty([2, 3, 4])
    expected[:] = range(4)
    npt.assert_array_equal(callarray(4), expected)

    # Test with Nones
    callarray[0, 0] = None
    expected[0, 0] = 0
    npt.assert_array_equal(callarray(4), expected)


@set_random_number_generator()
def test_multi_voxel_fit(rng):

    class SillyModel:

        @multi_voxel_fit
        def fit(self, data, mask=None):
            return SillyFit(model, data)

        def predict(self, S0):
            return np.ones(10) * S0

    class SillyFit:

        def __init__(self, model, data):
            self.model = model
            self.data = data

        model_attr = 2.

        def odf(self, sphere):
            return np.ones(len(sphere.phi))

        @property
        def directions(self):
            n = rng.integers(0, 10)
            return np.zeros((n, 3))

        def predict(self, S0):
            return np.ones(self.data.shape) * S0

    # Test the single voxel case
    model = SillyModel()
    single_voxel = np.zeros(64)
    fit = model.fit(single_voxel)
    npt.assert_equal(type(fit), SillyFit)

    # Test without a mask
    many_voxels = np.zeros((2, 3, 4, 64))
    fit = model.fit(many_voxels)
    expected = np.empty((2, 3, 4))
    expected[:] = 2.
    npt.assert_array_equal(fit.model_attr, expected)
    expected = np.ones((2, 3, 4, 12))
    npt.assert_array_equal(fit.odf(unit_icosahedron), expected)
    npt.assert_equal(fit.directions.shape, (2, 3, 4))
    S0 = 100.
    npt.assert_equal(fit.predict(S0=S0), np.ones(many_voxels.shape) * S0)

    # Test with a mask
    mask = np.zeros((3, 3, 3)).astype('bool')
    mask[0, 0] = 1
    mask[1, 1] = 1
    mask[2, 2] = 1
    data = np.zeros((3, 3, 3, 64))
    fit = model.fit(data, mask)
    expected = np.zeros((3, 3, 3))
    expected[0, 0] = 2
    expected[1, 1] = 2
    expected[2, 2] = 2
    npt.assert_array_equal(fit.model_attr, expected)
    odf = fit.odf(unit_icosahedron)
    npt.assert_equal(odf.shape, (3, 3, 3, 12))
    npt.assert_array_equal(odf[~mask], 0)
    npt.assert_array_equal(odf[mask], 1)
    predicted = np.zeros(data.shape)
    predicted[mask] = S0
    npt.assert_equal(fit.predict(S0=S0), predicted)

    # Test fit.shape
    npt.assert_equal(fit.shape, (3, 3, 3))

    # Test indexing into a fit
    npt.assert_equal(type(fit[0, 0, 0]), SillyFit)
    npt.assert_equal(fit[:2, :2, :2].shape, (2, 2, 2))
